#!/bin/bash
#-----------------------------------------------------------------------------
#
# nxlaunch
#
# NX Launcher enables you to launch NX Sessions to a remote Terminal Server
# either via SSH X11 forwarding or NXproxy. NX Launcher uses the NX 3 OSS
# components comprising an outstanding X-Window compression technology
# developed by NoMachine.
#
# The `nxlaunch' command runs NX Sessions in client/server mode or local mode.
# Running in client/server mode is ideal when using NX Sessions over slow
# links like modem, ISDN, ADSL/Cable/Fiber or WAN connections.
# Running in local mode is ideal when using NX Sessions over fast links like
# LAN connections or if you can not reach the Terminal Server via
# SSH directly but only via other hosts.
#
#-----------------------------------------------------------------------------
#
# Author: (c) 2004,2024 - Invoca Systems <simon.matter@invoca.ch>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#-----------------------------------------------------------------------------

declare -r VERSION="4.4.9"

# Treat unset variables and parameters other than the special
# parameters `@' and `*' as an error when performing parameter expansion.
set -o nounset

# Set proctitle by re-execing ourself
if (( ${PROCTITLE+1} )); then
  unset -v PROCTITLE
else
  # shellcheck disable=SC1007 # shellcheck doesn't understand this
  PROCTITLE= exec -a "NX:" "$BASH" "$0" "$@"
fi

# Print an error message to STDERR and exit from the current (sub)shell
exit_message() {
  local -i RETVAL=$1
  local PREFIX=
  if (( RETVAL )); then
    PREFIX="Error: "
  fi
  shift
  echo "${PREFIX}${*}" >&2
  exit $RETVAL
}

# Unable to run with BASH in POSIX mode
if shopt -qo posix; then
  exit_message 1 "unable to run with \`bash' in POSIX mode, exiting"
fi

# Check BASH version
declare -i MINMAJOR=2
if (( BASH_VERSINFO[0] < MINMAJOR )); then
  exit_message 1 "unable to run with \`bash' version < ${MINMAJOR}, exiting"
fi
unset -v MINMAJOR

# Detect basename and dirname, relative and absolute
if ! (( ${BASENAME+1} )); then
  declare TMP_NAME="$0"
  declare TMP_PWD="$PWD"
  while [[ -L "$TMP_NAME" ]]; do
    if [[ "$TMP_NAME" != "${TMP_NAME/\/}" ]]; then
      cd "${TMP_NAME%/*}" 2> /dev/null ||
        exit_message 1 "unable to change to directory ${TMP_NAME%/*}, exiting"
    fi
    TMP_NAME="$(readlink "${TMP_NAME##*/}")"
  done
  cd "$TMP_PWD" 2> /dev/null ||
    exit_message 1 "unable to change to directory ${TMP_PWD}, exiting"
  declare -r BASENAME="${TMP_NAME##*/}"
  unset -v TMP_NAME
  unset -v TMP_PWD

  if [[ "$0" != "${0/\/}" ]]; then
    declare -r LIBDIR="${0%/*}"
  else
    declare -r LIBDIR="."
  fi

  # shellcheck disable=SC2155 # impossible with readonly variables
  declare -r LIBDIRABS="$(cd "$LIBDIR" 2> /dev/null ||
    exit_message 1 "unable to change to directory ${LIBDIR}, exiting"
    echo "$PWD")"
fi

# Define possibly missing but required environment variables
if ! (( ${HOSTNAME+1} )); then
  declare HOSTNAME
  HOSTNAME="$(hostname 2> /dev/null)" ||
    HOSTNAME="$(uname -n)"
fi
if ! [[ "$HOSTNAME" ]]; then
  exit_message 1  "environment variable [\$HOSTNAME] is empty, exiting"
fi

if ! (( ${USER+1} )); then
  declare USER
  USER="$(id -un)"
fi
if ! [[ "$USER" ]]; then
  exit_message 1  "environment variable [\$USER] is empty, exiting"
fi

#-----------------------------------------------------------------------------

# Common options for NX agent
declare NXAGENTOPTS="-once -dpms -nopersistent"
# Enable auto DPI feature
declare -i NXAGENTAUTODPI=1
# Enable the magic pixel feature
declare -i NXAGENTMAGICPIXEL=1
# Enable the render extension
declare -i NXAGENTRENDER=1
# Enable the shared memory extension
declare -i NXAGENTSHMEM=1
# Enable the use of shared pixmaps
declare -i NXAGENTSHPIX=1
# Common options for NX proxy
declare NXPROXYOPTS="nodelay=1,timeout=0"

# Common options for NX agent/proxy can be overwritten by configuration
# files searched and applied in the order as listed in `$CONFIGFILES'.
declare -r CONFIGFILES="/etc/${BASENAME}.conf
/usr/local/etc/${BASENAME}.conf
${LIBDIR}/${BASENAME}.conf
${XDG_CONFIG_HOME:-${HOME}/.config}/${BASENAME}.conf
${PWD}/${BASENAME}.conf"
declare CONFIGAPPLIED=
declare CONFIGDATA=

# Global options
declare -r EXITSIGNALS="SIGHUP SIGINT SIGABRT SIGTERM"
# Log ID string which changes at runtime
declare LOGID="${BASENAME}[${$}]"
# NX agent/proxy default log files
declare -r DEFAULTLOG="/dev/null"
# Usage of `localhost' and `localdomain' can be tricky because it can mean
# different things like IP addresses `127.0.0.1' or `::1' to the OS and also
# NX agent/proxy have special handling for `localhost'. Additionally they're
# also used by `get_cookie()' and should not be changed without carefully
# considering possible side effects.
declare -r LOCALHOST="localhost"
declare -r LOCALDOMAIN="localdomain"
# Patterns where `^' is being replaced by the display number
declare -r NXX11SOCKET="/tmp/.X11-unix/X^"
declare -r NXX11LOCK="/tmp/.X^-lock"
declare -r CUTHERE="=====%<-----------"
declare -r SEPARATOR="|"
declare -r SSHX11FORWARD="-X"
declare -ri NXLISTENTCP=0
declare -ri NXBASEPORT=4000
declare -ri X11BASEPORT=6000
declare -ri REVBASEPORT=2000
declare -ri REXECBASEPORT=8000
declare -ri LOCBASEDISP=0
declare -ri REMBASEDISP=1000
declare -r SHORTDELAY="0.1"
# How many `$SHORTDELAY' per second
declare -ri SHORTPERSEC=10
# How long to wait for a process to terminate
declare -ri SERVICETIMEOUT=90
declare -ri AGENTSTARTUPTIMEOUT=40
declare -ri SSHTUNNELTIMEOUT=20
declare -ri AGENTREADYTIMEOUT=20
declare -ri PROXYSTARTUPTIMEOUT=10
declare -r REXECTIMEOUT="60"

# Locations of binaries and images
declare -r NXAGENTBIN="nxagent"
declare -r NXPROXYBIN="nxproxy"
declare -r BASHBIN="bash"
declare -r SSHBIN="ssh"
declare -r ZBIN="gzip"
declare -r DIALOGBINS="yad|Xdialog|zenity|kdialog"
declare -r ASKPASSBINS="/usr/libexec/openssh/gnome-ssh-askpass|gnome-ssh-askpass|/usr/libexec/openssh/ssh-askpass|ssh-askpass|/usr/libexec/openssh/qt4-ssh-askpass|qt4-ssh-askpass|/usr/libexec/openssh/qt5-ssh-askpass|qt5-ssh-askpass|/usr/libexec/openssh/qt6-ssh-askpass|qt6-ssh-askpass|ksshaskpass|lxqt-openssh-askpass|x11-ssh-askpass"
# Linux specific
read -rs LSOFBIN < <(command -v "/usr/sbin/lsof" "lsof")
readonly LSOFBIN
read -rs FPINGBIN < <(command -v "/usr/sbin/fping" "fping")
readonly FPINGBIN
# FreeBSD specific
read -rs GETOPTBIN < <(command -v "/usr/local/bin/getopt" "getopt")
readonly GETOPTBIN

#-----------------------------------------------------------------------------

# Runtime options modifiable by `get_options()'
declare -i USEGUI=0
declare NXAGENTLOG="$DEFAULTLOG"
declare NXPROXYLOG="$DEFAULTLOG"
declare NXLINK=
declare NXPACK=
declare NXLIMIT=
declare NXADDOPTS=
declare NXHOSTID=
declare NXCMD=
declare -i NXPROXYDISPLAY=0
declare NXLOCATION=
declare -i ASKUSER=0
declare -i ASKPASS=0
declare NXGEO=
declare NXKBTYPE=
declare -i NOLOCALE=0
declare -i NXROOTLESS=0
declare NXTITLE=
declare NXQUERY="$LOCALHOST"
# Logger flag to send to STDERR
declare LOGSTDERR="-s"
declare NXDEBUG=
declare RSSHPORT=
declare -i LSHELLPORT=22
declare SSHCIPHERS=
declare NXPRECMD=
declare NXPOSTCMD=
declare REMCMD=
declare -i ENABLEREXEC=0
declare -i NXUNENCRYPTED=0

#-----------------------------------------------------------------------------

# Get options
get_options() {
  local OPTS=
  if ! OPTS="$("$GETOPTBIN" -o "Ga:y:c:i:b:O:t:r:l:UPq:hvg:k:nd:p:s:C:x:RT:e:o:uXE:z" \
    -l "gui,nxagentlog:,nxproxylog:,compression:,imagepack:,bandwidth:,nxoptions:,hostid:,command:,location:,askuser,askpass,query:,quiet,help,version,geometry:,keyboard:,nolocale,debug:,port:,shellport:,sshciphers:,proxydisplay:,rootless,title:,precommand:,postcommand:,rexec:,enablerexec,unencrypted" \
    -n "$BASENAME" -- "$@")"; then
    help_message 1
  fi
  eval set -- "$OPTS"
  while :; do
    case "$1" in
      -G|--gui)
        USEGUI=1
        shift
        ;;
      -a|--nxagentlog)
        NXAGENTLOG="$2"
        shift 2
        ;;
      -y|--nxproxylog)
        NXPROXYLOG="$2"
        shift 2
        ;;
      -c|--compression)
        NXLINK="$2"
        shift 2
        ;;
      -i|--imagepack)
        NXPACK="$2"
        shift 2
        ;;
      -b|--bandwidth)
        NXLIMIT="$2"
        shift 2
        ;;
      -O|--nxoptions)
        NXADDOPTS="$2"
        shift 2
        ;;
      -t|--hostid)
        NXHOSTID="$2"
        shift 2
        ;;
      -r|--command)
        NXCMD="$2"
        shift 2
        ;;
      -x|--proxydisplay)
        NXPROXYDISPLAY=$2
        shift 2
        ;;
      -l|--location)
        NXLOCATION="$2"
        shift 2
        ;;
      -U|--askuser)
        ASKUSER=1
        shift
        ;;
      -P|--askpass)
        ASKPASS=1
        shift
        ;;
      -h|--help)
        help_message 0
        ;;
      -v|--version)
        exit_message 0 "$BASENAME $VERSION"
        ;;
      -g|--geometry)
        NXGEO="$2"
        shift 2
        ;;
      -k|--keyboard)
        NXKBTYPE="$2"
        shift 2
        ;;
      -n|--nolocale)
        NOLOCALE=1
        shift
        ;;
      -R|--rootless)
        NXROOTLESS=1
        shift
        ;;
      -T|--title)
        NXTITLE="$2"
        shift 2
        ;;
      -q|--query)
        NXQUERY="$2"
        shift 2
        ;;
      -z|--quiet)
        LOGSTDERR=
        shift
        ;;
      -d|--debug)
        NXDEBUG="$2"
        shift 2
        ;;
      -p|--port)
        RSSHPORT="$2"
        shift 2
        ;;
      -s|--shellport)
        LSHELLPORT=$2
        shift 2
        ;;
      -C|--sshciphers)
        SSHCIPHERS="$2"
        shift 2
        ;;
      -e|--precommand)
        NXPRECMD="$2"
        shift 2
        ;;
      -o|--postcommand)
        NXPOSTCMD="$2"
        shift 2
        ;;
      -X|--rexec)
        REMCMD="$2"
        shift 2
        ;;
      -E|--enablerexec)
        ENABLEREXEC=1
        shift
        ;;
      -u|--unencrypted)
        NXUNENCRYPTED=1
        shift
        ;;
      --)
        shift
        if (( $# )); then
          NXCMD="${NXCMD:+$NXCMD }${*}"
        fi
        break
        ;;
      *)
        exit_message 1 "\`getopt' parse error, exiting"
    esac
  done

  # Validate final configuration
  validate_config
}

# Help message
help_message() {
  local -i RETVAL=$1
  copy_stdio_append << EOF
$BASENAME version $VERSION
Copyright (c) 2004,2024 - Invoca Systems
This program may be freely redistributed under the terms of the GNU GPL

$BASENAME runs NX Sessions in client/server mode or local mode.
Running in client/server mode is ideal when using NX Sessions over slow
links like modem, ISDN, ADSL/Cable/Fiber or WAN connections.
Running in local mode is ideal when using NX Sessions over fast links like
LAN connections or if you can not reach the Terminal Server via
SSH directly but only via other hosts.

Usage: $BASENAME [OPTION]... [-- <COMMAND>]

Examples for client/server mode (forced by -l, --location option):
$BASENAME --geometry onescreen --compression WAN --sshciphers aes128-cbc \\
  --precommand "start_vpn -p work" --gui --askuser --askpass --location host
$BASENAME --enablerexec --shellport 23 --geometry 800x600+50+50 \\
  --command "xterm -bg green" -l user@host:777
$BASENAME --geometry 1280x1024 --keyboard pc105/ch --compression ADSL \\
  --bandwidth 64k --location user@host --title "RDP Session" -- rdesktop \\
  -a 16 -f -k de-ch -u user1 rdpserver

Examples for local mode:
$BASENAME --geometry allscreens --query xdmhost
$BASENAME -g 800x600+50+50 -r "xterm -bg green"
$BASENAME --geometry 1280x1024 --title "RDP Session" -- rdesktop \\
  -a 16 -f -k de-ch -u user1 rdpserver

Example for running remote commands in client/server mode:
$BASENAME --rexec "hostname -s"

If a long option shows an argument as mandatory, then it is mandatory
for the equivalent short option also. Similarly for optional arguments.
Using -- as an argument by itself treats the remaining arguments as commands.

Main operation mode:
  -c, --compression <MODEM|ISDN|ADSL|WAN|LAN>   NX compression level
  -i, --imagepack <packmethod>                  NX pack method for images, by
                                                default auto detection is used
  -b, --bandwidth <bitrate>                     NX rate limit in bit/s defined
                                                as \`n[b|k|m|g]', by default
                                                rate is unlimited
  -O, --nxoptions <options>                     Additional NX options for
                                                NX proxy, by default
                                                only \`${NXPROXYOPTS}' is
                                                used, see man nxproxy(1) for
                                                more options
  -t, --hostid <string>                         ID string to better identify
                                                connections on a server
  -r, --command <command>                       command to run on NX display,
                                                empty for XDMCP session
  -l, --location <[user@]host[:port]>           remote location to connect
  -U, --askuser                                 ask for username, by default
                                                auto detection is used
  -P, --askpass                                 ask for password, by default
                                                SSH is asking for passwords
                                                only if SSH Keys are not in use
  -q, --query <host>                            contact host for XDMCP, by
                                                default \`${LOCALHOST}' is used
  -g, --geometry <WxH[+X+Y]|\\                   NX geometry options, use
                  allscreens|onescreen>         <Ctrl><Alt><M> to minimize in
                                                allscreens/onescreen mode
  -k, --keyboard <clone|query|detect|\\          NX keyboard type, something
                  [model]/[layout]|...>         like \`pc105/ch', see man
                                                nxagent(1) for more options,
                                                use detect on old clients,
                                                by default \`clone' is used
  -n, --nolocale                                do not send locale information
                                                via SSH to remote host
  -R, --rootless                                enable rootless mode
  -T, --title <title>                           NX window title
  -d, --debug <v|vv|vvv>                        debug level for SSH and others
  -p, --port <port>                             SSH port on remote host, by
                                                default SSH is called
                                                without \`-p' option
  -s, --shellport <port>                        Shell port on \`${LOCALHOST}',
                                                by default \`22' is used
  -C, --sshciphers <spec>                       SSH ciphers specification for
                                                encrypting the session
  -x, --proxydisplay <display-ID>               NX display for NX proxy, empty
                                                to use SSH X11 forwarding
  -e, --precommand <command>                    command to run on local host
                                                before starting NX agent/proxy
  -o, --postcommand <command>                   command to run on local host
                                                after stopping NX agent/proxy
  -X, --rexec <command>                         command to run on the remote
                                                host trough the SSH tunnel
  -E, --enablerexec                             enable to run remote commands
                                                trough an SSH tunnel in
                                                trusted environments
  -u, --unencrypted                             enable unencrypted direct
                                                connection
  -G, --gui                                     use GUI elements where possible

Informative output:
  -y, --nxproxylog <logfile>                    log file for NX proxy, defined
                                                as \`${DEFAULTLOG}', by default
  -a, --nxagentlog <logfile>                    log file for NX agent, defined
                                                as \`${DEFAULTLOG}', by default
  -z, --quiet                                   print as little as possible
  -v, --version                                 show version information

Help options:
  -h, --help                                    show this help message
EOF

  exit $RETVAL
}

# Validate configuration options
validate_config() {
  # Check for invalid options
  if [[ "${NXDEBUG//v}" ]]; then
    exit_message 1 \
      "invalid option \`-d|--debug ${NXDEBUG}', run \`${BASENAME} --help' for more info"
  fi
  if [[ "$NXCMD" ]] && [[ "$NXQUERY" != "$LOCALHOST" ]]; then
    exit_message 1 \
      "unable to use \`-r|--command' and \`-q|--query' together, run \`${BASENAME} --help' for more info"
  fi
  if [[ "$NXLOCATION" ]] && [[ "$NXLOCATION" != "${NXLOCATION/:}" ]] && [[ "$RSSHPORT" ]]; then
    exit_message 1 \
      "unable to use \`-p|--port' and \`-l|--location' with [:port] together, run \`${BASENAME} --help' for more info"
  fi

  # Sanitize NX option strings
  NXADDOPTS="$(sanitize_list "," "$NXADDOPTS")"
  NXPROXYOPTS="$(sanitize_list "," "$NXPROXYOPTS")"
}

# Generate OS specific numerical signal variables like `$SIGKILL' (which is 9)
# and add a special `$SIGZERO' which can be used to check if a process exists.
#
# The variables are intended to be used with the `ctrl_proc()' function below
declare -i COUNT=31
for (( COUNT; COUNT; COUNT-- )); do
  eval "declare -ri SIG$(kill -l $COUNT)=${COUNT}"
done
unset -v COUNT
declare -ri SIGZERO=0

# Control a process by sending it signal(s) to influence its behaviour
# Usage: ctrl_proc PID(int) [NUMBER(int)] [SIGNAL(int)]...
#   (1) process id
#   (2) number of how many times to send each signal:
#            0: check only if process exists, returns 1 if process exists
#        [1-X]: send number of signal(s), returns 1 if process still exists
#       -[1-X]: send number of signal(s) and wait until the process exits
#       or we reach `$SERVICETIMEOUT', returns 1 if process still exists
#   (3)...(X) list of signal(s) to send, the default is `$SIGCONT $SIGTERM'
# To get the exit status code of owned processes, use `wait <PID>' after
# `ctrl_proc()' has successfully terminated a process.
ctrl_proc() {
  local -ri PROCID=$1
  local -i COUNT=0
  local -i WAIT=0
  local -i NUM=
  local -i SIG=
  local -i RETVAL=

  # Get `$COUNT' option if any
  if (( ${2+1} )); then
    COUNT=$2
  fi

  # Check if process exists
  kill -n $SIGZERO $PROCID 2> /dev/null
  RETVAL=$?

  # Return status if `$COUNT' is 0/unset or the process doesn't exist
  if ! (( COUNT )) || (( RETVAL )); then
    return $(( ! RETVAL ))
  elif (( COUNT < 0 )); then
    COUNT=$(( COUNT * -1 ))
    WAIT=1
  fi

  # Set default signals if none were supplied
  shift 2
  if ! (( $# )); then
    # shellcheck disable=SC2086 # doesn't apply to integer variables or lists
    set -- $SIGCONT $SIGTERM
  fi

  # Repeat to send `$COUNT' times
  for (( COUNT = COUNT * SHORTPERSEC; COUNT; COUNT-- )); do
    # Send signal(s) only every `${SHORTPERSEC}'eth time
    if ! (( COUNT % SHORTPERSEC )); then
      # Send sequence of signal(s) to the process
      for (( NUM = 1; NUM <= $#; NUM++ )); do
        eval "SIG=\$$NUM"
        kill -n $SIG $PROCID 2> /dev/null
        sleep "$SHORTDELAY"
      done
    else
      sleep "$SHORTDELAY"
    fi
    # Immediately return status if the process doesn't exist anymore
    kill -n $SIGZERO $PROCID 2> /dev/null || return 0
  done

  # Wait until the process exits if requested
  if (( WAIT )); then
    for (( COUNT = SERVICETIMEOUT * SHORTPERSEC; COUNT; COUNT-- )); do
      sleep "$SHORTDELAY"
      # Immediately return status if the process doesn't exist anymore
      kill -n $SIGZERO $PROCID 2> /dev/null || return 0
    done
  fi

  # If we came so far it means the process is still running
  return 1
}

# Log to SYSLOG and if configured to STDERR
syslog() {
  # If `$@' is empty, STDIN is logged
  # Linux `logger' (>= 2.26. ??) adds a prefix to every line, we remove it
  # plus also the <PROGNAME[PID]: > prefix on STDERR.
  local LOGL=
  logger $LOGSTDERR -t "$LOGID" -- "$@" \
    2> >(
      while read -rs LOGL; do
        echo "${LOGL#*"${BASENAME}"\[*\]: }"
      done >&2
    )
}

# Check that needed binaries exist
check_bin() {
  while (( $# )); do
    if ! command -v "$1" > /dev/null; then
      syslog "program [${1}] not found or not executable, exiting"
      return 1
    fi
    shift
  done
}

# Copy STDIN to STDOUT and append to a FILE if a filename is supplied
copy_stdio_append() {
  local LINE=
  local FILE=
  if (( ${1+1} )); then
    FILE="$1"
  fi
  # Duplicate some code here for better performance
  if [[ "$FILE" ]]; then
    if { exec 3<> "$FILE"; } 2> /dev/null; then
      while IFS= read -rs LINE; do
        echo "$LINE"
        echo "$LINE" >&3
      done
      exec 3>& -
    else
      echo "${FUNCNAME[0]}(): unable to open file \`${FILE}' for writing" >&2
      return 1
    fi
  else
    while IFS= read -rs LINE; do
      echo "$LINE"
    done
  fi
}

# Sanitize list of parameters from `,,ABC,,X=123,,' to `ABC,X=123'
sanitize_list() {
  local -r CHAR="$1"
  local OPTS="$2"
  while [[ "$OPTS" != "${OPTS/${CHAR}${CHAR}}" ]]; do
    OPTS="${OPTS//${CHAR}${CHAR}/${CHAR}}"
  done
  OPTS="${OPTS/#${CHAR}}"
  OPTS="${OPTS/%${CHAR}}"
  echo "$OPTS"
}

# Read lines from STDIN and write those matching a pattern to STDOUT. A `+'
# as the first char of the pattern, like `+pattern', is the default and means
# to select matching lines, a `-' instead inverts the sense of matching, to
# select non-matching lines. Pattern `test' matches anywhere in a line,
# `#test' matches at the beginning and `%test' matches at the end of a line.
match_lines() {
  local PATTERN="$1"
  local LINE=
  local -i INVERT=0
  if [[ "${PATTERN:0:1}" == "+" ]]; then
    INVERT=0
    PATTERN="${PATTERN:1}"
  elif [[ "${PATTERN:0:1}" == "-" ]]; then
    INVERT=1
    PATTERN="${PATTERN:1}"
  fi
  # Duplicate some code here for better performance
  if ! (( INVERT )); then
    # Write matching lines
    while read -rs LINE; do
      if [[ "$LINE" != "${LINE/${PATTERN}}" ]]; then
        echo "$LINE"
      fi
    done
  else
    # Write non-matching lines
    while read -rs LINE; do
      if [[ "$LINE" == "${LINE/${PATTERN}}" ]]; then
        echo "$LINE"
      fi
    done
  fi
}

log_version_config() {
  local CONFIGITEM=
  syslog "$BASENAME $VERSION starting on host [${HOSTNAME}]"
  if [[ "$NXHOSTID" ]]; then
    syslog "hostid [${NXHOSTID}]"
  fi
  if ! (( NXPROXYDISPLAY )); then
    while read -rs CONFIGITEM; do
      if [[ "$CONFIGITEM" ]]; then
        syslog "applied configuration file [${CONFIGITEM}]"
      fi
    done <<< "$CONFIGAPPLIED"
  fi
}

check_agent_option() {
  local -r OPT="$1"
  local OPTS=
  OPTS="$("$NXAGENTBIN" -help 2>&1)"
  [[ "$OPTS" != "${OPTS/${OPT}}" ]]
}

get_agent_version() {
  local VER=
  if ! VER="$("$NXAGENTBIN" -version 2>&1)"; then
    VER="NXAGENT - Version (unknown)"
  fi
  echo "$VER"
}

get_proxy_version() {
  "$NXPROXYBIN" -v 2>&1
}

# Check hostname resolution and roughly measure round-trip time (RTT) to host
check_remote_host() {
  local -r HOST="$1"
  local PINGOUT=
  local RTTMSG=
  local -i RETVAL=
  if ! [[ "$HOST" ]]; then
    syslog "hostname is empty, exiting"
    RETVAL=1
  elif [[ "$FPINGBIN" ]]; then
    PINGOUT="$("$FPINGBIN" -C 1 -q "$HOST" 2> /dev/null ; sleep "$SHORTDELAY" ; LC_ALL="C" "$FPINGBIN" -A -C 1 -q "$HOST" 2>&1)"
    RETVAL=$?
    if (( RETVAL < 2 )); then
      PINGOUT="${PINGOUT##* }"
      if [[ "$PINGOUT" != "-" ]]; then
        RTTMSG="${PINGOUT}ms"
        syslog "round-trip time (RTT) to host [${HOST}] is roughly [${RTTMSG}]"
      else
        RTTMSG="(RTT unknown)"
        syslog "unable to measure round-trip time (RTT) to host [${HOST}]"
      fi
      RETVAL=0
    else
      syslog "unable to resolve hostname [${HOST}], exiting"
      RETVAL=1
    fi
  else
    RTTMSG="(RTT not detected)"
    syslog "program [fping] not found, unable to measure round-trip time (RTT) to host [${HOST}]"
    RETVAL=0
  fi
  echo "$RTTMSG"
  return $RETVAL
}

# Generate random X display number in the range `X1-X999'
get_random_display() {
  local -ri BASE=$1
  echo $(( BASE + 1 + RANDOM % 999 ))
}

# Check local environment and X display
check_local_display() {
  if ! (( ${DISPLAY+1} )); then
    syslog "environment variable [\$DISPLAY] is not set, exiting"
    return 1
  elif ! [[ "$DISPLAY" ]]; then
    syslog "environment variable [\$DISPLAY] is empty, exiting"
    return 1
  fi
  export DISPLAY

  if ! xwininfo -root > /dev/null 2>&1; then
    syslog "unable to open display [${DISPLAY}], exiting"
    return 1
  fi
}

# Get local keyboard type, like `pc105/ch'
get_local_kbtype() {
  local _XKB_RULES_NAMES=
  eval "$(LC_ALL="C" xprop -root -notype -f _XKB_RULES_NAMES "8s" "=\$1/\$2\n" _XKB_RULES_NAMES 2> /dev/null)"
  if [[ "$_XKB_RULES_NAMES" ]]; then
    echo "$_XKB_RULES_NAMES"
  else
    syslog "unable to detect local keyboard type, omitting keyboard settings"
  fi
}

# Get matching X authorization cookie (MIT-MAGIC-COOKIE-1)
get_cookie() {
  # First we strip the hostname from `$DISPLAY' because SSH adds
  # `${LOCALHOST}[.${LOCALDOMAIN}]' to it and `xauth list' does not
  # like it. If it fails, we fall back to using the full `$DISPLAY'
  # to get the cookie. If it still fails, we try to get the cookie
  # for `${LOCALHOST}.${LOCALDOMAIN}', which is valid if X has
  # been started without network connection and the hostname has
  # changed later.
  local DISP=
  local XAUTH=
  local XCOOKIE=
  local MULTI=
  DISP="${DISPLAY##*:}"
  DISP=${DISP%%.*}
  # shellcheck disable=SC2086 # doesn't apply to integer variables or lists
  XAUTH="$(xauth -i list :${DISP})"
  if ! [[ "$XAUTH" ]]; then
    # shellcheck disable=SC2086 # doesn't apply to integer variables or lists
    XAUTH="$(xauth -i list $DISPLAY)"
    if ! [[ "$XAUTH" ]]; then
      # shellcheck disable=SC2086 # doesn't apply to integer variables or lists
      XAUTH="$(xauth -i list ${LOCALHOST}.${LOCALDOMAIN}/unix:${DISP})"
      if ! [[ "$XAUTH" ]]; then
        # shellcheck disable=SC2086 # doesn't apply to integer variables or lists
        XAUTH="$(xauth -i list ${LOCALHOST}.${LOCALDOMAIN}:${DISP})"
        if ! [[ "$XAUTH" ]]; then
          syslog "unable to acquire the X authorization cookie to use,"
          syslog "environment variable [\$DISPLAY] may be set incorrectly, exiting"
          return 1
        fi
      fi
    fi
  fi

  {
    read -rs _ _ XCOOKIE
    read -rs MULTI
  } <<< "$XAUTH"

  if [[ "$MULTI" ]]; then
    syslog "multiple X authorization cookies found, using the first one"
  fi
  echo "$XCOOKIE"
}

# Add cookie
add_cookie() {
  local -ri DISP=$1
  local -r XCOOKIE="$2"
  xauth add "${HOSTNAME}/unix:${DISP}" MIT-MAGIC-COOKIE-1 "$XCOOKIE"
  if (( NXLISTENTCP )); then
    xauth add "${HOSTNAME}:${DISP}" MIT-MAGIC-COOKIE-1 "$XCOOKIE"
  fi
}

# Remove cookie
remove_cookie() {
  local -ri DISP=$1
  xauth remove "${HOSTNAME}/unix:${DISP}"
  if (( NXLISTENTCP )); then
    xauth remove "${HOSTNAME}:${DISP}"
  fi
}

# Check whether a display number for NX agent is free
check_agent_display_free() {
  local -ri DISP=$1
  local -i LOCKEDPID=
  local HOST=
  local -i BASEPORT=
  local -i RETVAL=1

  # Check if a lock file and UNIX domain socket were not removed and clean them up
  if [[ -O "${NXX11LOCK/^/${DISP}}" ]]; then
    read -rs LOCKEDPID < "${NXX11LOCK/^/${DISP}}"
    if ctrl_proc $LOCKEDPID; then
      syslog "detected own lock for display [${DISP}] but pid [${LOCKEDPID}] does not exist"
      rm -f "${NXX11LOCK/^/${DISP}}" 2> /dev/null
      rm -f "${NXX11SOCKET/^/${DISP}}" 2> /dev/null
      if [[ -f "${NXX11LOCK/^/${DISP}}" ]]; then
        syslog "unable to remove lock file [${NXX11LOCK/^/${DISP}}]"
      elif [[ -f "${NXX11SOCKET/^/${DISP}}" ]]; then
        syslog "unable to remove socket [${NXX11SOCKET/^/${DISP}}]"
      else
        syslog "removed lock file [${NXX11LOCK/^/${DISP}}] and socket [${NXX11SOCKET/^/${DISP}}]"
      fi
    fi
  fi

  # Check for a free lock file and UNIX domain socket and free TCP port
  #
  # There is one case where no listener is used at all but we still check
  # for a free TCP port `$HOST : $BASEPORT + $DISP' to keep it simple.
  if (( NXUNENCRYPTED )); then
    read -rs _ _ HOST _ <<< "$SSH_CONNECTION"
  else
    HOST="$LOCALHOST"
  fi
  if (( NXPROXYDISPLAY )); then
    BASEPORT=$NXBASEPORT
  else
    BASEPORT=$X11BASEPORT
  fi
  # Not a very clean way here because we actually connect but I don't
  # know a better solution (parsing `netstat' or `ss' output seems
  # worse and `lsof' needs root in this case on newer Linux systems).
  if ! [[ -f "${NXX11LOCK/^/${DISP}}" ]] &&
     ! [[ -f "${NXX11SOCKET/^/${DISP}}" ]] &&
     ! echo -n 2> /dev/null > "/dev/tcp/${HOST}/$(( BASEPORT + DISP ))"; then
    RETVAL=0
  fi

  if (( RETVAL )); then
    syslog "NX display [${DISP}] is not free"
  fi
  return $RETVAL
}

# Find a free display number for NX proxy in the range `1001-1999'
get_free_proxy_display() {
  local -i DISP=
  local -i COUNT=100
  local -i RETVAL=1

  # Check for a free lock file and UNIX domain socket and free TCP port.
  for (( COUNT; COUNT; COUNT-- )); do
    DISP=$(get_random_display $REMBASEDISP)
    # Not a very clean way here because we actually connect but I don't
    # know a better solution (parsing `netstat' or `ss' output seems
    # worse and `lsof' needs root in this case on newer Linux systems).
    if ! [[ -f "${NXX11LOCK/^/${DISP}}" ]] &&
       ! [[ -f "${NXX11SOCKET/^/${DISP}}" ]] &&
       ( (( NXUNENCRYPTED )) ||
         ! echo -n 2> /dev/null > "/dev/tcp/${LOCALHOST}/$(( NXBASEPORT + DISP ))" ); then
      RETVAL=0
      break
    fi
  done

  if (( RETVAL )); then
    syslog "unable to get a free NX display number, exiting"
  else
    echo $DISP
  fi
  return $RETVAL
}

# Find a free display number for NX agent in the range `1-999'
get_free_agent_display() {
  local -i DISP=
  local -i COUNT=100
  local -i RETVAL=1
  for (( COUNT; COUNT; COUNT-- )); do
    DISP=$(get_random_display $LOCBASEDISP)
    if check_agent_display_free $DISP; then
      RETVAL=0
      break
    fi
  done
  if (( RETVAL )); then
    syslog "unable to get a free X display number, exiting"
  else
    echo $DISP
  fi
  return $RETVAL
}

# Check whether an NX agent is listening on a UNIX domain socket
check_agent_listen() {
  local -ri PROCPID=$1
  local -ri DISP=$2
  local -i COUNT=100
  local -i RETVAL=1
  for (( COUNT; COUNT; COUNT-- )); do
    if [[ -S "${NXX11SOCKET/^/${DISP}}" ]] &&
       [[ "$PROCPID" == "$(LC_ALL="C" $LSOFBIN -nlPt -p $PROCPID -a -U "${NXX11SOCKET/^/${DISP}}" 2> /dev/null)" ]]; then
      RETVAL=0
      break
    fi
    sleep "$SHORTDELAY"
  done
  return $RETVAL
}

# Send notification message to NX proxy
send_notification() {
  local -r DESC="$1"
  local -r MSG="$2"
  syslog "sending notification to NX proxy [${DESC} ${MSG}]"
  echo "$MSG"
}

# Run user defined system command
run_system() {
  local -r CMD="$1"
  local -r MSG="$2"
  local -i RETVAL=0
  syslog "$MSG [${CMD}]"
  eval "$CMD" > >(LOGID="${BASENAME}[${$}]: system" syslog) 2>&1
  RETVAL=$?
  if (( RETVAL )); then
    LOGID="${BASENAME}[${$}]: system" syslog "command [${CMD%% *}] exited with error, trying to continue"
  fi
  return $RETVAL
}

# Expand and verify integrity of a compressed and base64 encoded file
expand_file() {

  # Generate data strings with something like this:
  #   `"$ZBIN" -c < [FILE] | openssl enc -base64 | while read A; do read B; echo "${A}${B}"; done'

  # shellcheck disable=SC2034 # shellcheck doesn't evaluate the eval()
  local -r ICONDATA="icon.xpm
sha256=304c8406e652c3cb45615a9af62121cbc03fb7a1cead63690a483f0ff2afd952
H4sIAMSzyWAAA82c34vUMBDHn92/olzbfbjKnT8QFREKyoEPio+C+LAswgmyB3qCIP7vNu3uNu1MvpmZZLcOKBduN/108s1MOk3u+rL49PF9cXm9+nm/uf+2Lba3mx/F
ZbH7/X3za7e9fbe9273d3G8+fyleF39WF4+fvCjcv+7/i4eri+LBtvhwt/vqfr7qfi4fdXZz45qNaz7tbGi2rvmss6FZuubzzoZm5ZovOxuatWu+6Wxorl3zprf+os6a
Ray/9J5hGYI9xcCwHIKD6BmGxvrcdoA4Mpyd4EjRMyyGsIfwGEpgdb7L1mOve4g9g/tleRW2jH5qjp0yDOsWQLS5ENpZlzOGugEQVR6Eau7aIwP5AGNZJOEh1DzDySXh
efrgV8IAJVGmM7S0M8pwWkmMbh4lThmgJJpESVAx8AwnnKD12JF3MxOGA8jJRoPqkTAwuNQSRmNEmLiTZTiRJDg9hhlglLBKgtUjYDiBJHg9IobskmDiI8vgg+SWBBMf
WYaJ5ZXEeEstuYEwQ41ClVYSQT1iBigJ5WgAMWAGKAnqUWTTxZuGIVsaDwUnjoFM0DxRgizeEAOxLJLwhpT/CmbIMUGxHgUMGSQRRYgywMQhGQ0QH8UMMHHEEcbBDOsH
zoveYOKIIVRNHCHuhyRJxPUoZIBRAktCIAYZgz2NR+KjhgFGCXCDMFlqGaAkggMdXrwhhvAEBZII3qNIj4QhbIYJKtOjgkEvCaEeNQzamB1NlhaGNWCgMVusRx0DXEuQ
bo+/Eix1ZPOiN4Uk5HokDLHPAoiJxz09StK7hgEmDu9zomRpZMCLfQ5Vtu5VMcgkgR8mkhmQJJpqzil9KlXMC2fxxBF7mIgxCCw2GqrgZGRAadw5woCgZwCSqAx6tDEE
Y3Zp0aONIRQlWpMeCYPwm6wkGj8+6qpmBj/wkqj18TGJgZFEpU2WqQxUEpVVj3aGuSQmelS/cjIyTGO2+7LfThkL6wSt5zI9gyadtROE2djoypdWhnoSlYlGVbJM10PL
Rk7NaCTPi6bmFxWK0bAxlFEESa2KZZCCeHGyCi5r5JIw+WG8UAlWVuIoYWEYfd/AR0CpJAwMXiyILHJPxjB9qoclVKEk1AxTPUY2bRieswQgNVm1onpZI4LQ+mG+nyby
JCzqVsnAlcChJHLXH0JVJvTspa7DxCy0akWSEMRsYw1k5mLEEO9ZMy/aYL9pklD4AZXAk7Z2mWqDTKdQEhEIS42Uc25KlDDUivkskCAJKQONj3OzRwnhvJC8koGJA0lC
5gdRlQlKAqVxEYPwqd6axkUMcTEMBiWR9I5VUQK3TVABg/QVoTPkCNk7dxZEVeIxSSLuB13J0SKJKIO2xAOjhI1BXXI0RIkIg+YVIf0KNVYSkX1Blqon3KHNdYPnha3q
qZWEcJ+YqrYDJcFMLtl+OWWVRikJwKB9ReiZThKi/ZP6DbSqCRpmMJfA+xvQ7PYT7Kc17WXWLPZDftAkS9YU5zgCDKbgNOsZQIj2mVtfyfiGRmMiMZ4hSY8HE28TZxns
r2QmJpUENy9q9pMGE0qC8UMGPdKeqI1CYxhS4uPMZFGCMoyjmOEklEgShCExPs5MlMbnDLU9WfIQaDQqniGbHg8m2AAaPDuZ6ZCi5LRX6AxptsOaOEqUDIN/xjWX1fRc
bugsbWJMtJvPsBCEu7J3xnwhggPD4az9AgDHs/b/xd8cWAxi+NsLf1+t/gEXpCVjtEIAAA=="

  # shellcheck disable=SC2034 # shellcheck doesn't evaluate the eval()
  local -r IMAGEDATA="160x160.png
sha256=21d57f3267b6fac8e933d97f6b7a9be9b1630ef5c55bc527cee071ffd07cc5df
H4sIAO2NGGMAAwHlTxqwiVBORw0KGgoAAAANSUhEUgAAAKAAAACgCAYAAACLz2ctAAAABmJLR0QA/wD/AP+gvaeTAAAACXBIWXMAAA3XAAAN1wFCKJt4AAAAB3RJTUUH
5gkHDBoVSOTvowAAIABJREFUeNrtfXe8HFXZ//c5M7t7S25ueoOQQhq9JpAISLEAr9hewS6CDSkKSH1fIgQIXQFBRAUVRJGiCIIgRZTeO6EIBALpPbft7syc5/fHtHPO
nJndwBV/r97JZ7J7Z6ecmXnOU75PIyjL8SecMHvJ4qUHBYE/EmDYF7J+bbQQ55zDspkbny3/0qSencILMwAi48qU/SCyD4nUvxhEpA+S4hFxsk94rnicVDxGZRsnWwiO
K9DT07sxj3njF9rIF/kelpLrrhw9etSN55577oMZSpg37/RHzz//glluqdTsqJHzzqw3lN0v/6apqYdCmXMSERzHARGBSECIcHUcASEIRAQhnGS/cFu4D5EDxxFwHIIg
Ee4bnSs+jyABEgTHESAh4AgBQnRukV4vPq/jCON4io5zkvOH16ZkXxIEh8KxLVu+AnfeeRdc193o50jUPJeg94kAQYRatQ+HH/6tx+fNO21WMroTTjhh30t/fNntlUoF
zPaXu3EE2AwRvlcCDAlOXYRCBDGxhSvBcXSijAlFkJMQj+OkBEAJYTjGNiRETYgIKiKkmMiEEOk5BYGEk14vIeR4nIDjuAkRChGeVzgCt912O9auXdtPk/69P+/+oUFC
b283DvnqV/Y7//zz7hAAsGzZ8kPdUjkivlAIcqEcTH/M36+RIM3/PTxnY0HMysVDrkeReOToOyKRmT7gmGiJ1O+UXJcgDAJnjUOI5DoCm202HlOnbA4ppXI9RdyHF0nF
tjFx0jGE14hvRwiBdevWY+XKlQXPp/Hz1Pd778+7PxZmRkulFatWrTkUQPi0e3p7NiXLDfU3EWb3618itHFG+4vP6rEx8Ybno4SgzE9ZagFcF0M6O7HZxInYdrtto2fF
qFWr1muQojsS0gmSXBekECPB8z3ccsufNNHb+Fn+3yFCEKGvr3dTAHCjl0h5BMRcJI5TlZk5TzSwxt6z+3Eu+w/35YbigZkTsRsbEwk3TF48onuhjO6jjkcIUSjqiAhM
AmvWrsHKlavR09WViFLXLSXXzupX2kVyJGM41pcWvKyNo/lnaX+e+n7v/Xn3CydESHNuKFZE8a7vmQhR8ECKH0pzv+umpcZdSLEyQ+aTGCTQxDCDSIDB0fMgMFO4b8LB
BMirhjqf4+K1V1+FcES4nQhuyU3JLRbJpHJDyoyNONQ7Y8KtVMp44smnUS67TT2Ld/fM//VEGE9Mkcx0ogZE+F7EcSPRkM/+Vb202Bgx4BZT9DIZep+pF2b1Pe3FJdzU
fC6sXMMkOIDY0DPNaygwEBHhxht+j1LJafp5v7tn+a8Xx/HzcFM9BIoOpO4pFGW6EScsEl/GrOKUKzVlqSX/CZggXAKZiHCM4XfF0oxgERIivB9HgIlS7hXtE4txCDc5
Rjgi5E5CgBxKzyHccPwRVMMkAEGheCUBcpxwvMIBRHo9EIXfhUj3jX5f39WNDT094bEICTpfOL17qKspiIYRPc/3IGalzDUQNALUTXydCDu33Q2b7LY/WgZ1RNO0mSsT
WCUS24xizp+MBtcjy9Oj+CEl7CMFijXNy9THSBUBBAKDY+PAeDnJsSwVCZ/qlgxARlsDAB4pXM/gbMgBnmOOKQCs6a5ixmF7bDzKv1GoMxvjURmDPrHTd2iA7E28/1pP
F1Y+fheqrz5eOH9cG56mXqhl8BCMmr4V2oeOMN92/tThUNBYKYxZv2mN2NhKjMScDkko2xUFn1gh14yEM8Srcb+JF4NSwkzdNwSS0s5TyA7mEjXmWSb2VvUlRsmmNfis
ByNPbaMmpDMVSHkbfZLBWDL0zOhZuwYbXn4c1ZxbEJoINiyu+OEwcyiKnBIc1w1FTC7xsV0fsbHgXO5nI0BW3Fw6p6OI01JmUhjczDLBqAFgq/2eoxPZCYqtnLiIeAUI
UgZwhHkO42XTP4cjNr+9WWiXEy9Sczpg4U6kEnbkWyXrzIogtFTismV2MOv3mLwvtt53otwzGb9TtCk0MBDBLOp740RnJSvphBYxJ8cnlBRzcErUYCCwUwCn9olBWenx
yS2rupeyrduTOudWny/baIIy4vK9EWGj7dyYOBUewBIAi+JhmUYIFEswj7JlRDAUEwxS6zNjrMZPnXM4nzTPT9EFLISdI4FCsUsGt1AmjOJdMPkaR2NPIVDSKIo0Xcgm
tnUppRMMWYmHLe/WkxFtZ54VNRCd75P/FvkTuPAlERcyUaEbIapksxChacyw/sS4GV+RuY0tJMVFbJ1sky2LZWh6ERWYiZQj1sj+sknYJ0gzr83EURXuV4u533vmWP0o
gZuAagq5JRsMpmCJdEBFVacs12JwZAmKrDHbaMBsmSG2F89sIdxID40RZMuNs/J/OvMoNZPVG9NnUQ4EpHNuEqSoBzmQgmjyLRuH13y7K7GhdUDvkh7zRAk3UAUL7t1+
ETbj7xrhgEK/JqWKHKtAtAKncxHE0gzns0IvZEXMn7jyLHh9vVa9XAgH5LiFRoZ6kTHbzMaE3fY3FFbdm0Im8yQbq6Bizt2EZ6IeSB0toI1jX2/edwvefui2jHrglMqY
dfg5cFvaCmGWkDGoD5UL5gBtFGdkLnZuFBohKSekFB2J9WRmyGbtJObmuDoXz/0pHz4Ifzpy336ROi/fehX2v+BmjJi+Q6oXWmAu1YZNvpMDKJAMN4ikY6vyEL6XaqDq
n7BQfbHZKwMfD/3gSOtv+5xxHdxKm2VSN9LlGmA2nCMtkr9CNyYLByi1gSEavg+ha1WZON2E+7Kh7xV6fziyWNg4QFqOkYoIUFbm0LplJgwePxXTD/hqv6k+fz3j0MSo
YlOlja4ZgumGxKP4WRhIEqvbCYz0HKG1Hm2LaMyXQF2G7ngm4/dkJdtjAQOo9/Xg+s/OsN7bjE9+E6O2mZ177Matxrgo1WSYWb9vEKjUCum2goUb0UsRBxQpAYbOcIuy
T9EgEkLh8K1JpKs6Yhn9buJ86r7mcQX2fPIQmLDT176PIZO26BcCrK5fjbtP+0roFtMsZNJcygQVHYhWIdL9mRLXWkJEKtZNKkQV3gdLQp/HgAy/s0TBav/9iZ/OhfTr
WYW+pR3bf3Wudi2WlP1bu3ajMVjW5L0w4JSAUguo1Joqw9xYLYmls8gV89r0jj6kZZZIgCXrM0LdX5slxnHm/kza36oKABD2/N8r+40LLnnyb3j9nt8nHC+5phK6pXG3
ZF5F+6uwZoZ7q/ejSxbJCuySO/nIkEHp+txvL8Cbf7sxc2R50BB88pdP61KGdLsgCx/Qu7eshQPR0gE4ZQCO9iys9mSRCCZBBhHqI1X19ZT9RmvEbDdqZRubJ+1vGH9L
AC0jxmH2sRf3GxE+eOHRqPV2JaIw5oj6mEjBGtV9svegEkqeCO3xZKI/ZldkxZ2yrn/nH3jpDz+23svOh50Np6VVm7OZ88DyW564ta4MlFqASjtQqkDa3n2TiWWx902o
NiPlWT7MWeDYXM3tMt9CT8SX9WmZTy4CjaM7m7jHJzBmhz36jQhvOXyvDLJhWraUURcMTsXKPsZKyn3XAyR6YQGTU2AjnYJu/87eYBlk7mGnb56F8XMOsFOtjQIp6/HL
tYKZQeSA3Aqo0hFG97BhMJngO6lKVB6vJ7sRkqG9hGtxKjbRhEi1iK9ESeciZTf9LkGQivIe/777/1wBp9LaLwTYt3oZnv3tDzWRmeH4CueWIEA4msOnUIlXdMSqz+l9
ZJ6JsRqGyL2nftY6/sGbTcfkj3xpI4wKyzuCXQViBqjSBi63gp1SqmYB9vNk3js1qQMKspsBZPhvbUaIzLdkNW4mozVXNigcL9GdDNdU9LvrtmC/C//Sb1zw2Wt/iOUv
PGoYRqkVq01QG7dn+77qfn11Tv3WeZwpZ33x+ouw4vkHrWPf98J7Qouyaf0n55qKkUnCBVXaQZVBIZSiSa48icV2qdgABxRAGKKeiwWRAZdsjIsmq2jkbCMFfyML7pmK
OYpmaceYiZi638H9RoT3nnlINMeaCDthpC/dFKeZfQmSCYHERuMgLBnrFr6IF393gfXUe595E/oJb4kCUF1Q6yDAbc0aKLaJp3l4yALJcZMwTMbuUIAJhuYNSUUt51u8
Nqs2V9SSppSn2CMlmJrKlznRIAgzv3kmOsZO7BcCrPesx52nHBTmhWiWuMUiiOEZIgVmUcVQem9EQG9dNmn16qv0Pdx1wn7WI7f49HcwYsasLExE1IR1q6ZQMOCWQK3t
4HJLNG7OUZ8KjEnjmXGTzEpknJlkseElg6VCZTHeJ6WO/Ul1RRNKYbjKyMJKretI+Y1W7VjFP83M+ODcX/cbF1z23AN49Y5fa2POWP2J9Z9u196Egt0QM6qetB+fd95o
lUGAe+d+OgxrN5bh03fG1l84IX1m6j8NErNdO90GpwTR0g44ZWWy8XteOfnMJ8JY6opYHpPFAFGB2fQ5cxNjoAwWZvtbMzIMDpngbYnRQgkozpT+PWjMJOxy1IX9RoSP
XHYCulctVcZRZPQbEErmWVEUcGAaGGSHWpRtL/3xUqx57ZnM+Eptg7H796/Nh7fIzpU0DuVWQK2DwaVK+PzNfUiHXhrSZg5XLOSEqg6ohbZbiJAawSu5/kSyejhi74Gp
K5Ki56WOVMqEUZEGUAMT9zwQY7bvP2jmtu/uA+kHWfFrQi9CWKiSEhWhr852aZsHZ0XbFj/2F7x47XlW03Gvs26BW27NnxUSWV9hxJVFpQXUMghwyxGXKvDN5+nrufvq
8pisgU6c8aILW5QEGQ5mqTF5ZM3xIpPf+C6JDKPZDsFwFJWT5ThZABuCsNvJv+w3Aqx3r8MTV57awFCl7HiU8foM+OAGx9vP+ehFR1jHtc2X56Jjk6kbZ+QKAZTKoI5O
SFHSxphrIxadu+F+3ASfShOeXHskVJwtZnA+KgyGyAm/IQXZIGVimjUDUr9shjtqqUXKAJQQdqfUio9edC/+cvRe/UKEr9x2JTbZ+UMYt+NeIfeAhCM9VGQVZenBZQ/E
kQ5MBEkCvighIBc10YIen8Bw7eFbBfFXfzp0K8h6NpVnzE4fxtSPfStxBxZFrsdYgmhpgxROuCXgpiORcu0GI57Sau1y/jnSpPf0d9cKuahEGLF1VlPrWYO707gyNadB
ecBJSHua5Z0lUPM3RibVUosxI2gwCBMwZLMZGD/nALz90J/6hQjvO+8b+Ny1r+HoBcdiVP0dCGI4HMChAIIi55sIxySdMMdXkgtfuJAkonxggZpowcPte+Cx1tlY5oxF
QPaqBy/+7lx43eusv8058Zf5lKESHhGo1AK45ZALRQYTyHIA22CTnPwPLbDYkhME1oMQlAiqTGgBp5iea1XhDC7FkfXLksGUBs5lErYkK8Gbae5sXJAnPbc9lJ/UmUWU
3ZcUfsrR21dycRnA7ON/hrXfno3uZW++ZwL0+7px+/H7Ysz8K/C/z30NEEqiVvJBWOcOxxttW4AEocx1OLKKMurolBvQKdehE+vwye7r8ameG8L7EsAL5W3xWmk6Vovh
+Htlbyx54k688odLsq/DcfHRSx5qEP8W/uVUWiGdEtSKDRoBNcrPzkRuKwSZ4X7WRKCQLpIo+iZFcEpqaX6sRsmspEOw7iLOTIZMkDApgZkFBoqWuJP9jRQqZDXeiY0g
CgbmHH8l7vzePv3CBVf/4xn8/e67MGj3c/Hdf5yopgeiThWcNv1X2CAGJxpIXUrUffusbpU92LN6L3b1HsBk/3VsEbyEN91JuIfm4PEf2fW+aR//NlqHjcstq0EAqNIK
OK7O8ZoRrdyYo9qPNbiicb7mKnsoEdFJfROk2WIZNSOJzqRMLgFTnl5YnKytaUScoxcJNb0gdiBS9gSKCjB00lbY9eif4JGLvt0vRPj4z07G6G3vx4LBs7BVV5rpP2/q
FdjgDAU4SDSMWt3MEaGEo/eJQbi9/QDczgcAANq5CyQIdx2+B4JaX+a6E/b4DLb87Ik51EGgUjmMw2OZyBvKoy019bUZfc8aRFK0Tx7Hy0lnjQYozEHk5ulwQbSLCmIB
mWw0irZpMXJJ3koRVySdMGNIhsnQM0mDc8p+FVvusidGbTWn3yzjO47bF5dtfga6nUEACC90zMSG8kgkQD4DNZ+TQkjJrGTkplP0UAce+PmZ6Fu9JHO9SucIzDpkLi73
vo3p8hWUZTXBM0SlDaJtMKhUSV+EGZFjLrIA+rGJatPNYd2vCSgHxc4YNy9UhpVKB3EkTFIXiJomdN2/m1EzqQn8MCW+NA9YF7kqm/3Uixdjztt/QuCU0X3qLbj+oE36
hQD9ag8e/tF3cfZ3rsD8Zz+PId4qNT0Ekgn1gHUctUHyzFt/vw4L77Z7cj7yo0cxrrQcnUE3TsN89FA7Hhaz8YuWI7MiLlet43xx21D3KxLdRWKbFNGVTxwxfQkbTmzT
1zjH0Z4PRFPuuGPMSyrfM/pgQmciwa6KEtDnvPVHXHLzLtjnjWvR6vego74Wcxbfhr3PvLnfuOAb916H5x55GL8cfzw2rb6JiuxJ9NEeT+puWatbMxU/1XUr8PTlx1qv
s+NhF8IpVfAWNsPnnKtxuXMY3hSb40N8D66tfhofDu6w0oLOdLggUIUbBAk32NdmQAsHJJyk1mGYO8INfcIiEwFiEiGFURkgw4eJHN9oPGxOP9Pt6Q2FaLnuu0pcfcQK
MMuKJcxpUgyFVvmUlU/hkj/ugi88c45SfCj8+MKCczF82kyM2GKXfiPChy48HI8M3h2PD/kgjlt4HACGH0gN8c+urH0CwN3H7GY9/5idPoLxe3wGLANQuQxq78DfW/fF
ma2n46uV36CLBuHQ4Ge41vtvbCmfB2TsJpDK85aJP5akEsoUJ/fYokj0uCwkqZBS2Q8yKRRFhDA/Jq4XpPmbg3D8TXBkoQtJWyiUit9QflApp3aKnm0Wndmqo6hxs2qe
BeklM9QAzejcLbVuHPXgUTjmgcMNU5y02LW5D30Ze51+Ezo2mdIvBBjU+/Dn7+6JX2xyIirch8HeGvR5sgnPUDqhHr/4MPjVnsy5W4aNwS7HXgk4LkTHUHCpBYw0OqdK
rfh6+Sr8wDkeIOD7fDrm4nS0yy49IklhxBKsFx7L888WsLkI1gZxiGtC9ctLg1tKGdIqioqVshENY3gkTCJkGH4Y0/BI/WxpMpHUw5Qkk5ZopCUCseIhYbXYkErY6XV2
XnQnLvjThzBj5eOF5cYWD9ocZ8z5NcDAzG9f1G9csGvJ63jppkswf/PLcOKbR0cXbC4odNWLD2PJo7daz7v7abeA2gaDKu3WZKd4fYx2wWfpBjyOnbE1FuAXztexF+5N
StlRzLxkGtEDBkhGurzJAI0Qu/Q9RwU01VA4MwtS4UAcRJxXcsYXxzm1bQRgq7OkexsyHE+a4dxKlIoWO5ZavqC8MPBsDF0S7WJyPgbO+PMBOOTxufbIHSWAcd6cX2P+
nF8lYx42bSfMPKL/EpqeuWY+lixahGuGfxNfXfXjpiKV1r7+DB6af6D1fLNO/A3aNttC99WqIlymKkosFc/j43EizwcY+LbzM/ykdKQy4dkIZk+PAxv+fOUdQEsjMKKg
NGnHWpKaDGQaKYX0+nlEyBGbTIMRKI8IzYI+KjxiwCSs415xEUkVRAZbzkkGfBPDNIqlO2HtSzj71v0wrG9FdiIpCTXL2zbDGbOvxvKOiRnv/IQ9DsLQSdv2GxHedfJ+
eKztAwjIgcNew/2fvPRw6/axu34cY3c9IGsJ2DwOUq9X+IbcHF/zf4ZFclOMoNX4SeU7mEyvJ3Mx9lDFECApFmgSt0NpSF4m7N4morWomyCKWTRjQPP9weotCe3FW4iw
KGElow9qYlWZaclvnCQ0a5xP1S9hHgOM2vAWTrr7YHRWVxuTgrQ6OMvbxuP02VdjyaDJhgGQrvucc0e/EWBQ68WTlxyGq4YdhnOXHaHcvxl4KvHX43ZD74q3MucYPHEb
zDz+19nAUSlTrpdsM84d/b2eO3Csdx7WYihGijU4r/VUjMESI8BXz22G9n5VI8IIYrXmfKdBszLOC9fGFn6nIjCQMzogwVrelmGpcMWwpoVBj1rOVj8lY3/Vp8x6DGBE
+TPfvB2n3fGZfBwqGu/TY/bEabtdAync4spMDHzghP6Lol7yyM146+5rMG/kuZhU/4cFz5BYdO9v0bNsofX4Xf/3xtTXqUaVm+mG0qAiiUwexNf7LsGTfsjhfzzoJOxZ
egCq3I2L/SeaSmQlkxKFnnU2ZBJVgCAAB0E0DqXkhTquJvOGhAoKarqU5mI1QEJTDCsek6RZjFqFm00s0iJ2DauHAExcvQCHPjY3g1OaZS+WDpqAn25/RuIDI1B+ji4I
Y3f8EIZtvn2/EeGL18zF6pqThD6pN9K9bCGe+8WJdqPj7HvQMnR0vr/VNrk5xwcb/XZW9Ti8GkwGCPhO688x2X1LEa/QUhvUGpJk+peREqpuaARpoIPJbFRGxVzoFskA
0ZnSRAl9pElDWihUPGxSQBy1sgAhJ7iUtO9aUUlBiYjefOUzOPGeg/Nx7tib0LkFTt/zd2kCe0F2P5J8Y2Cv+bejffTEfiFA6dXw12N3xVvuJO0BMxgPnfkp6zHj5nwK
Q6fuhFwcxBqRko+nqUef1Hsa1sghYAAXdJ6GLdxX9IBiQ0DJCEVUvVSZqhcsIYMgqyaBs/szI8vY7MENwhZdoRMhJ77bVGdTEriZEp+uXs+FLPmkZHxXfMTR8cTAsN4V
OO6vX0/ZnaUAJwAEwsFPZv0gDJ+PYBDOxbSy+unMwy/pNy7oda/Fyzeclyb8AHjk3C+ivmF1Zt9Bm0zDTsf8MqvPqattu6pjJTV5FJhF4Vpf33AxNnAHwMCZnedgBK22
Ys2IYBtSy/CpWDQjErnSKEAFJUdc/eQkpdQeFq2H4onigIg0zY/JTBLSuaMWZp5XY8XcnsA3KVE6QR3H331Ioa8zpsdT9v4j1pWHRyY/ZcPNE5cfpbCVsg6bujO2++pZ
/UaEr91yMda9+Tyo3IKFf70Gq1+8P/tE3RJ2mXtTVrFXuGY28Subicdx7IuSM2MaoGf3fDd5XucMmY8S6obayDlpl1FYvZSQvh8+Xza4nlTGYcAyMfjNVABGq56QJNKY
LP29FQxPS1RKkpWMGh4mh4MFptHkvh7T99knz8fQ3hUNi7c/PWZPrG0dAyl9Rb+BtVJVUWzklI8civZRE/qFAJ2WdrRvOgOrX3kcC351snWfOfP+jJahYxWl35LaKQ3X
khbREvWXS+At02hJ11e8KbipLyzsOdxZh0MHXZc/q6UOPsogSFNCpUGlUjleKqmz0XchI2VHFtmCOeFYNiJUWX2c3aTNBjZqyLCxXYNj9O3JSwAwY+nD2P31PzQu3Ang
8pnnh5BBIDUugQhwTdOV1TxVPWc1Pm6fs//aLwQ488TrwqSis/7b+vvmnzwGnZvvGF1fQjKH3EUaqwmJ6AhyyoESqIZz83+v6jkIK4OhAIB92/+G7cvP5eQLp8cFvp+e
WyrnU66X5IUzhx4XGUfNy/DeJBsRVQVWcJbL5CQBRaCl1hqBKcvhKAsqq/uz5tZJM/oPfLpBfm80rl9vdxKYCEG9FvooRaoWWCtgGNHdqf4a/uhW2rDLd694T8Q37bOn
YNj02Xho7kch69ng0o7xW2D6507RGYpiOdoT1xVUSqs0m/dGFQGq7H9Dz8eSPQ7tvN4KMjMYUgYIknRUpTIuq604lEmgMSSVycDK/WxDFrlMxiDCpF6yUr6WLUECyXay
6ISc5gNT5CWJ99v27b9hk/WvNXjN4R08uul/gQM/Gj5paIA1d1Cz+sgIWgrHNm7mf2HkVru9K+IbMmVnbP6JY/DSb09Fz5JXsw+53IoPnP13yCD1xaYBpFzc90WFVqWq
h+UXBEiQkEjfu6N3T9Q5DP3crLwMu7Y+lR4fGTMy4npQ9ExCiBMmRg6zrh6oIlkRw0bN40Ii1KxgGxHGlqlW1cBKfKT7b9UqANp3JJCJ6u341kMnZF2Blti6O6YcjLrT
Bul5mqdG5nlrjGgeq287Wj9w8g2oDBm10QS467y/oLp2GRbeeqldNJ90o1K+VlcHMtwwU2UVBR4Ji6GSCTAI13PXHpbMvRNGXJHi2UGAIPCViCvWiiPFalLi7VA/Ne+M
vr89aUl9BGwA0USwtjeg1II1g0Gtn2pUJumBDfYS6MD0lY/DYd/Oillv+HzzjMMgvVrK/Sh3+jRUJC1Jg5h11E+bpzwSmH3G3aitXYZ7j9jKruOUWzF0+i65kKytYIK+
c1F2mcVpy/Z9XqtPRE2WAAIcBNim8hKk7yVBASGMY4mgZhPbAsykJFL0HELK5anoaWs4oJFZZrbIYKGCyRaoJan2DjsQTUoVA4oLeaf1UXZ/4yY00+r2rc4ZCIQLKSWY
CJJ0MWpn+mgY0atCH8OnzcL0Tx7TFP1t+sEvYvDk7fHMpd/IB6nrfXj8nM/Y0yvyxtdE+eNMedxMnRb9+NVyCNYGnaHqQ8BHBj2g9/SDCfMga2gawbWpbqgUjUpSh7kw
lVSzgm2JbFo4v6QCYBnZyjqG8ZHUe1END6VU7y6LmgsQWNIxGbLupVE5/VJTgjPQx4xPfQ9tIzYtHMvIHfbFVl+7EE9d8CWsffmhwn1XP/83bFj4PEhGfJuzLi4TyM0f
KyfgsbXrQMH6aN+2yYvdq+MJkDTuXwOYDbA5iieMP0naxh6eR8Su4aIoBFUEFxkg+jQly3fKBJKaxIdc4gO2XPpwsbRUvr82dLuoC7coJC5qRHwxTqU+POVFEBPmnHqH
pQNTNCTHxTbf/DFWP/93rHy6ucnz8Cl7we9ZbxC9WcquEeEZgQIyn0DNQpfS83HPhpnamLZvfVm/f3MCyDisXyc+vS1HapxQ/HvAYQdvbqwSiUaQhxJHoXsxyBS1unEC
00vC9oI8uy66rXCMpIxnYfvmYHIUuqLcGumZmshIxUVGpCjoPhwBdsvo2bA6srSzy94hqRjdAAAgAElEQVQ/XQgG44nz/nvjglgv+4auuKNRkR+OsELF0MoYGFnoRqVF
6QdJAMFrVR1w37vjUeVZKMdLVgwKRQRLxder4pIx5NJ0M8f0rQndt2EnQqsGbeN8RV4PbVsqssev/0fBNRW1mBnL28Zr1EZoJHZM5z5nEP/wwYXygkplsHBR9xmPnb5/
Dph8PJxSCx6d99GNtpZXPfdXvHHbpZnZolvBWRzQ7P/HBWZUqjbJxMhgqRsK8TKpZYnBfZVIGSiwi1SMDGmI/xirjIlPNmM82YBoiwGSEIFWWFEtOEmZFlVpUUujKKUC
2cRQiJAeykG1KYsVROh1B+dCLYUFE60O/TTwE6IEuGECUCAZD535Xwhq2aShzsk7YfInTsALP/8Oepe9/q4ww1evm4eud17RJF4cuCslRwGe+goDbjGJVM/vYASejyAI
NDEdc97uoDV5vBWqg9jXAhtikWwGmcaDTcelFqtnPYghFtfghg1r9AKVuURo5odyagVTygHDAkFmrJjerZmV8CEhfZRUArR1TDezoyit0MR5eQeW1gFqqiiUiBUqlQHH
icpbSLz0+/Ox/rXHLXBKG3Y8/kasfPZuLHng2vfkNXn0zP2TmotgGbnGdOhFLVOcutakbqmq0c4y9GRI3w/vJbFOlUQhMKQircrkwYGPMN0ySlqCUogq+k4q0UdjJom0
ja/klNgy4fjFRJiU6M3jhGFCkYkzUaZec2JamwA1saKLsdKfLbwxx2y8UlRxHkr6o3p9ZSVSnftm7msojgAJEg7gVrSQjdVvPI/X/3i+9co7nfgHeN1r8MxFn3/PPmO/
bwNe+NkRSa4Gq6I3qSgmjWlv6FmKG08GPmTgh01sktrWMtXPWCY6bouT9phzKYBgGQUzx4QKxQuiqwRJnnFE3IlBgth4k2oapM6+MhV6lSr5dud/yhW1hneq50PphaYZ
GbacEcUTofYCkSSafnFDe5dnKrGzEsemKec2J72MQnrcSmjMSJmIpuraFXjwdHtF+mlfOAuDJ+6Ata880m+hW0sfugHVdSsKipWnc0eanQji6HcpEXiRC02yknNu8VZE
38vk625ISZphFCMD6rFpDxHoYlqxxvV9bSqqToRaoxpBFrZD+W2fVUOBCrGTnDJr8aQRApKcYseFcsC43jftaqKtipcZwcocwipuKevyA/DwBZ8DWzpQDt1iD4zf+2sA
gLFzDsS4Pb7Ub0R4//e2Q717dS5MliSZG8nkHHG9sG2XmavDlh4e4fcWVLXn5rOT6KBJXLtxLBmZecRKfLlaGFP1fEjkhmKlARakc0CyscIoqoUFGZ6O1O2W6oLICUpN
vSQqrMME+I6LmttabHwoYnnz9S/kVKM3gjlTECMS+wBKlbCRMmXhjqd/fhS6Fr2QuWypYwR2PO6GsBxcdD/TPn8GnEp7/1AgS7x89Qn57bPSQieJ/hxIGRKf2kYCrIfB
wSyhEn7/QOdz2uWrsgyfRULEpl6peTbiv2ECC5yJRM/v0MWZgIRUBywiQgmtWjzlNhmkbEunTDM/pbkLHHiiXIj8qGPfrPs1GNEHiYKc51Ug4YZ9MHK8BRsWvYh3Hrgu
R++7GSbFOuU2zD67/0Txiidvxdt3/dyasU9KBApHeh6C0ACIuSKZPVkswDJFYPInht2vvV9fOmAplP1jMFmxZBMQOgakYWxP/46/Cy5CiSIWRDYg2kKEZDRgIRPXM6w3
XUEmhUiQ7eXJhAUjZzWKEEiWqeufi7wYFleapjBF/N+thIIl2sZG85x61xrcN3dv67VmfPl8tI+eEte4UFZGpWMkJn3s2H4jwld/NxfVNYv1ZkAxNwoCyMALczIiDE4r
PBQrvxJJUGj8PXH5SUYJdYwurdGu+9SGacqx0flkqkxTZMCQQWCkEDjFx6mNhYJwf24WBySjU5JKhAljYwV6yWRYsSGGFdGhxqlR9rj7J3yioQSOfxvsrceY3jft0b8y
rg4lQ1jFKRmJ3dCjkJnx2IVftF6qdeQEjNv9y3qHUGOd9IkT0DpyQr8R4ZPnf0rv0BkEoXUrdcufISPcTUYqRro96Y4UOWKTBHdmjHLXYGi5W7vmHat2NqLSFfEr08pD
rFl9UgOsNZUx1v2IG/cL1nrFEeW7wBTohSzNBTXKMUJNyFZwXV0JWDx4Cny1anyDvonHv3Cc7h4wEH5yy2A4+bMvOuyZnx2BdW88lfm5ZdimmD3/0YbhTxxIzJx7D8gp
9QsBVlctwqvXngIEEhzreFavjfKADR92/J1kVqn87Kh7tOt50sHbfaO1oNhYtGckmeEXzvjdYxhGDaFr2C0zA0TbrWAmymn5ZpTU0DwjpIe/w9LGC2kE81+mfLnJUD7G
yOoSbNb1igK1hDOchAOmUpioL3W90AxMXfHcPVjyyB+sl9ry65dnS1OonoHYOwGCU27HtM/1X1bdO/f+Ar0rF1lTMzXvhOKrVfM2UnhEh15aqQ8fHf6Ydq3fL9sjPScr
xynuN1Y8HSncorgNpZ7GyUYKrAr4ZwlQs4Ipl+WQUc9P63PGuuPfyilzzHAV7n5i3IfCRtBoZBCHHusD3r5GCf0mQJTCvrYqn5LQG/lFT662fgWevNgOpWx9+NUYPHmn
bAdQayxheM6xu38R4z741X4jwsdO3wtBvU+LmIlVGTIbHSqgNElFImjfge9s+nvtgUomPLBm6yjaPQWU1TAxSgqIKlwPSlkPmUq5NC85woWl4btgLo6GIaFmZWYJISQ2
JZEI2UQjIw9GiZQxOmhTWu8vHvjrnVtifXlYtqSHlRAZu66+FxO7Xw4rMTilTDSSvbdteM2H59uDDIZtvQ9GbPNhS5BxTodPpej35E+cDKfS1i8EKP0aXvnNCVpgKCnh
+GbFrDhjTp90Kfcb5q7HPsOe0nSrtd4gvNI1PqkJY6p4UEL7YVbdkEqxfVYrrhBY2gIm7ERI9maFKhGahGgrtmKGa+Thcaydg0j5hYCgrxtzZ1+Ry6YzSikBx7w6F64M
UjdTNkY4MYeIGIIIz155OKprFlsVkm2PuBrWgDsjhyOxQBWu6rS0Y855z/cjNHMzVj7zZ2hUAMOlqPyWGGLQIxiIGddufXrmdR75/FGhqy+2WpXzkSFPWWk+k1s104w8
ypSlz0Icug5IVBi5pbnPMnAVN1T4zXGq5wz6usEgrGkZg+eHzSwAMXXOOKK2Al9749w0Fo2zfmFO8CaBpY/9EUsfvcl65l1Of9hS+zpCXaDGaJqTKiV7cisY/5Ej+o0I
F/zicNS7Vhm5v2nUCmkBpCkWqFqxh465FYJYE5VPrpuCVfXBWlXUNEBXcbWpPEZaMjkzVXFVuiVdEmnqF+cFpOpcUIdizLIaOZ9m5pwlMFStlCoDL6ysSQIMwlUzjjGs
p7xkm/B8u6/6Cz667AYdGeC0Nx2J8PYCr4pnf25vWjPpk/+DyrDxKW4rOYVq1IbQiriSUQVSLamcGRM/fhIqQ8f1GxG+eMW3kntNS2EoBgP0/BF1Em7f/g98fuy9Gcvz
xwsPiHRHval0YsgoerMeAaMn9MNWOVWRO3l6fFI7kMjkgEYvB0XqUUavonfR69NYgwCyrwcgJ/SuAFjcPgk/3OYswxEcV95ia8DEwW/9CFO6XzSCKUXorZGAX+vD/Sfv
an3Bo2b+N8Z/6Ah7eQzNqwI9Z0MagZoKa9/++FtBbqVfCHDDm0/htd/P0zwbpEEvrEAoKTccQl24YJqe3ccEnPrSF7GoZ1QCTocGhkwBbOYUY5QyDQFTsNPkU3WOCUAK
ghQEjvz7cSJaI2RDZCMAKd/hn5f8gmJCzKiPkiH7ugFyM2rXEyN2x8ud21i4oD4YtUPXGS99G5O7X45sZKHZDa/87vuorV+euX+3bQimfuF83feZITYgB4PKbou8D+X2
4Zh64Px+44JLHrwGfasW6ZigNhkUl5hkjHTW4sZtTzOkH+HFDRPw0KrpYVuHIGyjABlAJnGGYdicFAJwSqH70i0DpRaQWwG55bDTulMCOeUQ/xThSuSCyAGRA5ATSR4H
jVNlFSs426NLDTJt7CrLi0jWnNARfib7ukPYRLkwy7Ti0+nbX4Yed1AhHGN+m//KYZja/YImkl7/84VY/MBvLIcL7Pz9B0GinNEZM14eGD1RpFldnjU8kpkxapfPYMyc
L/ZbwMLjZ+9tNYZiw4E5zPsgWcO1W8/PPLZuvwVHLzgC0m0BuaUwCNetAG4LqNQCKrWGROaUIISbwCJhASo1RCuiBxV6yWNEDfyqWsd0gkiKVWfun1LiyeStstmOPYug
mxUKZLU3LFxDQukdGxe5lon76Hszf2e/BQuzjhWIea8egX1XXB/6eTesxMLb7K0ZNj9wPtzWDmSbtphcThqi1iyiJ5VSuToWNOmAkwoiKzZ+eeX6kwHpAYEPll74nf2Q
+ADM6nwZd+34P3qh2Uir+uqzJ4Dj1rsw+oQkrkyzUhenkdGcFIPLzblWPTNaZ9UcIoz180xAKpmd0xVfYPYdqTMReQVLUpjAryPwagrxGXkIccS0cLGuNBSf3/2BpFtT
Iz9x/PNXlvwYB6+8Eg+dvAM4yFauH7njJzB2zhft6YyxM15Ll5R6nWZzjYlPCVaAZDjldsye/2y/EeDyJ36P9W89AypVIEotgBuu5LTg62PvwFmTr8gIMCJgn0cvwFqv
I5NgRKx7eDIl7YwAX1jKyaVBETBq3rC9F12uK87whBCZ2hYVOpdtbt5MUpSUkLUeQLjKBXRwg4SjGeYM4LCZN6V/UOPQQQDYb+k1ePCzEzCyNRvsOvVLF1pEK/RYOmuh
SFM8NxbZotyGsbt9pd+I8LmfH4J699oE1G2nXlwxeT4+P/JuKwR24DOnJkGk6v2BTY+UZruCcgqCsuU1ZB0QeXTHBZ4Qy5tVa4jbc3qpcTEChRH6fV1g4Wr5IWmRmpDr
SaZMfuva8gh8ZZc7sbY8PIf4bDVfGJM7S1h62DR8YYvORO+bNf85gJxsDi0sZTCsRGdE+9iIk/V6yZM+cQpaR07qH3XQr+Pla78HZmBm2wu4ZfoJmNi6LPMUVnsd2P+p
+VjtdaReEhgJWzaHEbNVh7f+zcW59BK2UiTZUtA5QHRKhGS2ZsipQso2cRa7rfq6UoNDhQ+kBMEJDRK10mbsh4ysu7pow2E73YwFHduno+PGYnltLcCP9h6Nd749FQd9
Yx7c1sH5eZwmzGKzejUIxrI9CADpA4EPeDWgXgUFPrb6xq8g+gmacZc9gj9NORpnj7/c6KccccmuyTjomVNR9SspNBOJXhONIKlb/tSgwgQZFRmoAPEg5f2wpT8dCbLo
gBbOQoClkUzBOzThvsTocJJmfYkVSU4ushOXW4MSkXPaFpfi+k0OTThzZs6QWSwzCrFyBH7a+hucvvxk7ND3RDZXOK8ZC/TkoAwwHcgoI82DDOqQ0oeUQXifpRagpQNc
akN52GaY8rkL3hPhzdliCH53/HZYcd2H0eb4uq8+skB/ueSjOPqVw9NgCQmtkoGW8C6LOZ0tvz8vNtJ8dmq+Mxq45Nw0xsTe+zcJhEiKRMel3DhrbmksmgGvBpZ+2EuW
o6CxmJmSSCo1MbOhaxAEIczFUEJoGMDvN/kqHh22B8556Vsow7PDhZYScgGA7atPY4f6M1gvOnH0yIuwzulEnSqGFkuanIm1X5kUBApCy13K8Kk5Tgg8m5M3rIuRug63
2x/rXv47VjxxY9NE11ISmDCyBbecsgOmTRgMrxbAq/pqzy2AAA8uDltwLN6qjtGYRhIpowmxmA0ILcI9q2KTZsSGlCCQX5chzfcgzVenk0lS5S/64qp/hPF/uvVAiuun
sDWlrtWCvRpQrwGOG9KdiEx9ivC/xK1kMi1SipfbO0283TIJB297G/ZbfRMOXnZ5tv0c605GUNS9MyLoTt6AX646FKvEcCyobI2LO45AFa1pZGXc6DSpzxxln5EAhIAQ
JcB1FFelUbhbS85P72DyZ+Y3RYCdbS4uPGQa9tx6OCaOboHPQL0aKPcYRp6UhpZx8YJ9cfPqD8BnF/aSVHqkkhKUlZgfpJNQFFbFGSMzfTbWICVNB8+LEYi5pEaA+sWy
RGhtXFekf/keuF4LrdpYVkqAyFUK26gn4qT7EsWVFmSxC8cnF7eOOAh39o7DcW+ciF3GtMLLPBGyePXSgY+Ua/DB2v34YP1+rHJG4FctX8ASMQYLaTyqqIQvQzigUktE
0VLhbhaCswU9aimOAjudfD+ePHt3Xa9zCNtP7MDUMa2Y+5nJ2GLKYPh9AQIGvKgScZw1QQQ4LuG+V9bh82c+hbGfPwTto127ScY6iVEhpA9LUck88UkRJ1XenUksXCiD
IaN4NhcAAqM6AUWRqJyxu1lPLLJ1GfG9sEi3cJOXFUekJCmSGhGH5yQhFPacUzfZMu3uuuhLuBvAqFYHfz9wM4xuc+HDDCnjjK0ft5+ITzyCV+P46qWQQsCHizXOcDxY
2hW/dQ5EH8o60GWb+mrvPMqBICRQGTwa43Y/FKse/iVOOmACPj1rFKaNawMTwXVCcql1eYCgUFGPtHlmoOQSlqytY+axD2PZ2iqYgZVXfgUzj/kL3JbOAmJia5ynLYeS
M6AbWY5WClPa8rC1nBA7t9IIUEob6VNyIs3KVU1QaQxKSshaHyAcZKthqS/KwB0jVLxRFhVryc8Bnrzo48kpl/cFmHH1Quw1vg2H7jQB+44NFFOMjIdLhtGi+iYlyuRh
LC/HZ7yb8ZngFoAIa2gIFjoT8Q6NRS/a0S3a0c1tqFEFvWiFBxc+XAghUYKPQaIHFXgYJHrRQT1oFXWMd5dhc/cdjPpaH1qO2htcl/Bk6NhPKuELpGWOGXBdguis4Le3
v41f3f0O7npGT2T3qxvw8o0nYKsvXa49W7LqaMr9Svv2vOMSjsfFRgWyMbNWIgwCjQNK68ugQsyXMz5LWe0Kdb7Yhyic/AFTOMMJOe10bIWKlG3LH78RPYsXZA679+1e
rPncj3Hp8M3w/ZXzsE39ebRRtXGwRYPfh9M6DOdnsTOejSAESoiFYoNJUAoxRNtiuAHRNiICB4S+7uj3+DzGfRIBfYHEX55Yg4POf6ZwYq79x/1Y/NBV2HT2wTnBG6ph
AItI5gbovlIPxvprKo7TKkGyUG8LVA4YpjSS5aVTmpfOedQcBWZWuyOxK1KIkdkay6cWMmeLeKREZ1SbM6fVFVY/fydevf4k66Oa/o1rUB46HjIIcNqw/0VFVjHRX4QP
1+7Fx6p3NKPk2H3OXKBEkY1oVVXFHtltGuuVFgcY7OLS6xbimvuW4oVF3eipBU0Nd+Gd52OT2QcnIAWrxMZ5elxWtCato5l0hINSUssyIdYbEzIn7bqQw4V1HTCQSfds
0wCJAyAT6uYsBCOrXZrYJRJ6+QaRnd76jGR70JU2lNDxH3hV/OPG/7G+hKHb7IeOCTuDAz/krqKEulvGq6Wt8Wrr1vgxjsIO9adxYO+N6OQNmMSLINQa+w2A7fydLIp+
k4RerUu8sbIPy7s8/PDWRbj7+TXv0k3CeOSc2dj1xIfAkqzhbHlV69mYMMRmp0tOquJRSFq2ZgYKN2QtAp6yOb4WEWxawQaaIGWq/2m+wWpXGFSqNBxgaelDRhTBL+mo
TPxPB1btDOWJM3ZFUO3OPP/KsM0w5XMXpxlW2uRJl6dK2+Gpzu0h2IcLH2PlcnzIuxf/Vb8Tg9Bn5X4qqm9+0rvgpN3cilt7dsM9PTvhNW88njh9WwRNl7ctAB+qG/D2
A7/A+DmH2nibdWYkOISBCSKXwOzkq4pqiajwZwEQzYFmhMiMFcyKdZdW5tTxIq51h0GIIHAchWxIXT3JiQ2kgjTpxpEFHGfPCQOpeeuO863EBwDTv/U7yCisTItHI7uX
PCAHARy8KTbDFZWv4IryV0AsMY6XYQK/jfG8BO3UhxG0Gp3oxgheiyG8AR3oRUtc4iypdRiOt8Yl9HA71mIwVnlDsB6DsZo70cNtWOSPxkJvHJbK4REInD6IrQ77I567
7JNotst40fLWPT/E4AkzMXjcNhlNT1os3AQB01A++/xhJTyFDe7L2os1iyplTYyAA4MAM1a6wgktTZO5FnK+cDihL5dtXCHK9zCJgCidl+oNsZLTqtJR16Knsfhv9iYy
0w69GuVBo2A1pbmgjjKbc5mwmMZgMY2BtcVmYMP31Eg743paLZ1subR4aR0xCSO2+zhWPXtzv/iLX7r2MOx01F1wym1G8xkk5oI5prx6QmzitTYrmU0+y5lm1SYRSo0D
BnYNmwzfbTyPuNYbiV1HbwrLxuwSIgeQtAsuvSlfWnBE+jWsevZWjJ71uYz+UR41DR2TZysVUy3RChkDymadqZiWrUO2vWt5KL6k5WUAhX1i42Tz6EyTDzgdwinD613b
zJNqoA9KrH7lrxi51cdybNaNOltWiLMN59Ob0EhkqyKoRCij3xIdkDIUEF0srroUyXSu96TwipqhrClGkUkjDUekqmWQBRwmhavE6YSiBKfiYtIB3888rJrPCf6YdQsS
su4U1oFom2aTablO2Wx3TuV6yFGkVj0ChoFlH0KWE03c7xSD0EmxVRn2UBEDbmEF+4xqSueAMnmAmmUrWQ1GJNCLpY2XVLdTxhUX6DpgkMOG4mtIQAZgrxojpQrXMB5M
AjlIhe5Yc/gnVnK8D0XEEvmKSThKE2r7i5RmsBnDiKSUxRBXHCpCNuJUuak0RKcucihDH7bqrMj3jmhxENxA9df1NjVYItZ/UoDFFiqQonUNMsEz42KF3Gw+Ws6MVxb6
bmOa04yQ5BkpRBjnaMh6b5T1REkJrqznBHrvME0aslVpJYo4GImwmCTMBGZYZ7/nc0ErMs4Sn0RjF5qNgGQOcZpSIofTFeXnhJ0CilkR522RZjvd9GZlgfNAlQ1kmZ+s
8BxTVZCatqyTH2lXYAScZ29bPSGWyIeYCGUQRra0tIMTzsXZLutkN+LZxvajQDwSYRpfTJypLmU1v5JL1APOjbQtlCa51MBWhboosAAaBBt5fhohMdzQ2WM9kqy3mKJy
pJkabOOZ2je27Mm5DTfZegxbdENtT2LYY5nIYoTIwO4PJIpKw9YR1GrIFg6kTJSAGjJEGcd3TKsUEl7gawnnlBN4YM5mz5dN0V0xL6Ecy5RzOF1x90fKUqb15TdnRMBy
3uLMLMq9Vg5sbgYkgZt4bqqVm5QkMnhwaHwEXi1pgG07X8YTosbgqa/H93rQvfpNuOUV2Ruz+Y8p622khPIc/aFSSqRJkVY2CVEn7poXZMt9MZow5bhYXBfKTzTJq5Bf
TcQCrpPNTbkRXhT7OAj8HvDEpvz+Nk5rsFC/3guv1p37Liy+4KwRQgDcZU8BSx5HwGzXKLh4GxFBiNALQkRwhAMSYcRzGIIlQEJAiDA7WTjhNiFI+SQI4UAGEqvfWpiE
d6ncinNgDi7gYhowq6U0c4Y4NIGScDe20LLFR8C2tvWcTwLUiAZtTcCLyYioOeCFmvUlEjXo8SxSNMRyqkw4FhXNKOEWiyAjxU8L5RYi8o6lxBZHwoAIwnFAlBJbXFRI
OKFXQ4iQcIXjYPnKpRBO2TrDiS0gKbM9vEjp1s3GNjKtPbaIWIO1UYb2dCkS/xcX5aEmiRBNEmEzvIybJELeSCJ8N/xUJUAR44Da7ppRoYhTogxx2pw3lHmAIpkJFIcv
RUQXc8nYZUdxVHS8PbpuEATo6elJSgdTMw+DbNzEEpECgh/4WLlyBYgE1q1dC2Jg/fp1AAEbNqwPJwEJdHV15SZv2SIRSPlv1OjRGNTRERFi0QviJirMc6Gjx2o/N6lS
cBOlNWKEpNkxmrtqBMgKDpgHWFLOi6Y8XRBRvJ9KTArBpamfobdfkBqmpeh+EREuWvQWhBDaW6VmxII1HD1LhBMmTMRtt9+F3t5u/OjSy9HT24tTTpkHr17Dd485Dr7v
o1ar4cgjvwMwI2CJCRMmYNrUaejp6cbUadPh1+uYvPnmGDJ0CDoHd2L8+M3AzKhXq9h2u+0xYsRIMEtMn7EFfN//P0KE6BciNM+pwzCB1IRPagXrHgMbTpjulvWiCEqD
MJNtQo2EJ6j8LOWQKnEC3d3dUYqmCZCT1brNqOHR2Chjt6Z/tba04rrrfosPfeSjqLS0oK21FQ8+cB/+dt8jmDVzO3R2DoHv1TGoYzAAQuDVMGvX2ahUKhg+YgTmnjoP
u+y8LebNm4+/3nsXWlvaMH3Gljjky5/DA488hVO/fzLGjB2Lc879AR5++AGsWLYM69avDQNUi6JGmJoqMWNz+DdnoeX/Hp6zsYWn1vtrdI20RbPCAfWQfFtc27vlhJTE
YsaExZFIjY0LinvSkfI3RWXWAAghsHLVytz61WTNdcgX0WQR0fF31y1hxx13xprVqyAlY8yYMbj++t9h6tTpmH/WefjgnvsgCMJevLVqDTvtNBO3//lW7DxzFnw/iGZ2
gJaWNvzsZ5fj2t9ejanTp+PlVxbg/vv+hhXLl+LII7+Fa666Cvsf8PHwuQMNX14xk+EmuBY34G6NOCH3OyeM7z3SAQOrzpini1KRykkmhBJbwbE2qIrZ7IGq+knCwdJl
S5W00X8CEUZ/bdiwHq+//g/85MeXYPCQTtS8GjabMAmXXPRD7LDDTvjWN76Ku+6+A2+/vQgf2G13zNhyS3R3dWHffffHO4vfwUsLXsSnPn0Qlq1YjpdeehEsAyxesgRd
3d3YsG499txrH4AJPd3d2NC9AW+89poBMf07EGHzUFgg09Qx7LHnRx5etOjNXdOCMFmUOx9msEMxmgUrRAi9CBGaI8IJv4uIE0Ymu+M4IGI40e9SMt5+550MuMY5uB5b
sT6by848AycGTxAEKJVK0aQkyCCA67rwfR8Mhuu6ISePSnMEUkIIij6dKKc2QBAEICGS3OJSqYQgakATRukLBIG/kZykkTimBgZqI3gmn700bRmjscrADEydPuORh87m
/ZgAAA9aSURBVB+4d7YLAMJxl+k6HrRoDM7xHSUaYyaZnTSjOX65ItdaRcglobRwEgJLly3JXji5XHYwWsd2RU/M6oTQQqHS7ushIYbEFiGFBPi+n9xX/F2l4CCQUdVh
X58rMb5KgOd5OgfgwB59RFRAhI10Ql3XImqs79lD5qlAx2ysUzbWWxmlUmVJqgMyXxn4vsV9BrveZLMxjQsKEmDNoECSQUaUwi0Eoe8TEeGGri54nmd3L9HGiuN8K5gK
7X+yXJpy0NJ8vYU2Qpf+54tjfheic+PgmUZjDAIf7R1DrkwI8OgTT79986lbPtHb2x3NYr19PJvFJoFM3wy1b63W0FghqpD4HMUAoQRfE46AECIRv+vWrksrtMPoEMRq
FXdkyuxm2zWYBXqM8bKt365xfpjFODlTY1Bv+qePDUbvXmZLDzGz2lNB8TO2VwnKNBS0d2OXWklhZg4LSBX8ru1nlCTOu4aUxjUCH77vY/YH9nnimJNOvSMz7T/2qS+d
uvD1V/avVXvbGvv+isVAYrWqLXXM2oMxlK0mhkT6lSbqsrpBQ/WZCmKbuPhoWamUt0VOxx8w1tbqtcXW8zIHnUM6t6tWq9YLS8nrvLr3ToZT5akmzMXMRlBjTsiNQhia
clw0ha4WX4LQ1j6od8aWO9x60/VXnpF7yWPOvEhUl68UWR5KjW+gsVCwuWStmfgh6Gzz7NO7HAUatnAkAJddcqZ/7PdOzE0kGdQ+6OpV6+pfs/162Y/O9C/44Y94yeJ3
8kZ1dc0rZ471c4dLxRNdNBcQJJDJfX/PCzclivWn29HZGZx7xnHaQa6524WnHB2X+fuPXaZMnZz71Mvlsjz99EPyaAajRo5EW2slb1bIww/7lo+BJZ8ABxZgzKjRzSAd
1qWjYxAGtbcOPMQBAnz3S0dHe+5vsoFsKFfKKLuiCSh4YBkgwDydSQh7vSQC6p7fdvL3zxunguwy6rPxo59fuy4pTbKxNsMAAQ4sKaE5uUDqoEHl//6vj+728SAIIIMg
8YAIR2C/D81hP/CTfIeBZYAA36WFl8+qfD9wfD9wYoWQyAnzXP0AAYIGiuKAEB4gwCaWQMr3lFeRb78MyOABAmxi8Xx/gFkNEOC/bvG9oHkH6cYplwMPd4AAm+GAQRMB
lu+G/gYIcIAAm1gGDWrbiMiPdDdW/OA5Za8HBPsAATZexoweFTXVidp4Udw7F+k2IIm8QRIxkzatYUvhprpXHzHwdAcIsHD5w01/mlGt1ZQQKmjf47CqmMjiujASaUhW
cozB8Gq1+qyBJzxAgIVLe0fHJzyvnnA0UjqAhzFwOueTUIgzrhLLnDQ3VJdABqMHnrC+iIFHoC8B+5O9egCv5qNe91Gre/A8D/Wah5rno+55qNej3zwf9XqAuhft7wXw
ot88P7syMw4/8siBZz7AAQssVQmnVq8C4LQIJqfN+tIm20j1RDVCOreFVWiG7LrLB0qXXXppbeBJDxCgnQNKKVgzKmIxzFrmHWu9U9R0BL2jgLnIQLoABghwgACzy7Mv
vbj1m6+/PSmslRN1e+e4dUTU002mHC7JCZEhd2NIpbKshbsS4JRLP2Dmw4loIGIB+P/XOXnpT68YNXrUmBme5w2BZCd/tE3kRRib1LK0jqBS+6C2fUHOIZ7nwffqiShN
CUlqFq7K5ZIkqKSxdXxdsg9HMAYP6kS57PYJElctX77yz6Vyya3W6mntG+4fT7Qgko7jrFu7bs0/vnP4N5cMEGCTyzkXXPSxLbfY6k9CxLVjEKV4Qin7pibRs711FCmF
zpW8KCJA+hI9vT0gEprXQ3LkhovLBgdRLhpBaVmmY4JJHzXZqFgQUK/XMHTo8LRLFBFYSrS1t6G3t5p4S/rTE8NgfmfRou8ed+yRlwwQYIPlvB/+aPzUqdMWCSLmsHt1
WL4oTtimNJNOrfxu3gkbhBkbByXHxfoNG+B5fvSy9WrKYfnCyLJgSro+clIV2NAFlVTQOME9ocCkgJJ6CYZbaslcV0qJlpYK/CDodwIMCV3gnXfennzisUctHNABC5ax
Y8d9WQYBZJJcL8DEEV5ERlNPSuoLmV0ckyKQlPbLIBJYsWZFUhjHJiKFUlLOaCKWXlWZDGo9RALATlShISorT1FyHxFDRn30PK9uvXx3t4dNNx2HJUuXh5XF+nnpHNx5
KIC5AwSYs0ybNKziuO58zw+UogRBVN6Xk/KdZMzsNPlV7y3CyfeQIBxieF5Yt4VznLWu66bgKMUlKZQKCVrPXU4qeyXNcWRK/HENGVI4YSAlPM+eGCeZMXL4cCx88204
Tv+/GuE4pxz8xYMvuOo3V60fIEDL8r1TLn6MpYQPmb40iogQBvHFZT+ipneJODWMDLVPTF+thh132AYPP/pkVCDS1koecITewpQi0am1NBWkETolnFFXbsxyKn4QWAnQ
93yMGTMKLAsmyHsWw8Q7zNntoat+c9VWAzqgsfzwkstHdnYOWeGkb1+rkqoONqy+avRSthkhRtOfvp5uTJkyCSNGjMKbby3C8wteQrlU0uotlFwXwolKCsfxLUrpBltz
e0qaacd/CyUJPuaUIaf2Im+KKsprtRo+8qG9MG3KJFz+0yvQOWxUVMyo/xdmxpo1q8fNO+WEpQMcUFkcIW4CSwS+1Ov7abX+lOrwKopGAsgpzaYaJaVKCx548CGMGDEc
W2+1LT642wewoWsD1qxZi2XLVmDdhvVoa21FyS0pNQ3JELhRAAIp1J9YtKk6QEZVcCIgCEJiq9XrGDViBNxSCS2trZgyZRIC38d1N9yISksrPK+OIAj+aRynpaVyM4BZ
AxwwWk457azdN91ss/vSpsl6eS/KNMIx2yjZS76Z3S8ZjEHtbVi1YgUWvPgidtxpJ0yZOgXDhw3FyBEj4AiBV197HQvffAuLF4cMIq5tyFHzPeK0/3aqInAIE0VlyURU
iAkkwlqBEii5DsaMGQUJgfb2QViyfEUIXUuJ1SuWYM2q1dhu223hM2Pduq5/avQ0AVi1auWHzz3z+3cPcEAAQ4YMOcOv10OcNkRQk65Kuh5mK29mm0u2llqRpdnVjSHD
hmO3PT6IarWKZUuXo2tDF1atWoXhw4Zi3NhRKJVLGDp8OFatWZdw5EBKCCKU3BIcERcYFGF7sYgivSCA5/uo172w4GZkojuui8D3sXjFGgRBAEHL4UQ5xZ5Xx/DhozBx
wiRUa3WsW78BMiqV989c2trb5gG4+z+eAx71vZO2mTxx0nNh1VAKOYmi6yUcJjFCqKA5S9wyKgeYjvy1ggTckotyuYyWchmlcgmuW4JwQk5Xq9WwfkMXunp6wmKLgqLO
tGEjwNhAkHGltkBCRqXLSHBYVCiOFuSwO6jjAI5w4Dpusj2eFIIINa8Or15Pwex/dvA0EZYvXTLzsosveOI/mgNuMnaTB2LYhWIWyGljUQgR6fOcFpBWOjkViioVjknA
5ZBIa/U6enp64TgOypUyWistaGutQLhuCAYzoVQqhwA0M4JI9joqkUfQEIuomlPCuUUSf+9E+B9FxF/3PIAlgkBCBj6kDEIHi5Q6bPQ+PPthw0f+FcDg/1gO+D+nnrV/
+6D2WynGVwzMjYwKpYQEYNuoBB8GNP+Ymu4R+32JAEeEnFEQQcq0eDtHwDIhjXROJG8cKQPd+yKjXr6pLRSFcskAvh8o0dX6J+ifk5CXt1T7+j5+4Xln/Ok/kgMyBzcE
vkd64UqzfG5OtyPNKClkgFlBrHhI1P08MFzpwhVOQmSsEBii0Ps03C9112nkHYPQ6iAjIvODMDhVL2PLUR8U+b4SX6hGyD8AKP3HccAjjznxnGHDhp8oWWrYmdoWjLRG
bSYWGDsXmijJrv6piWIdsQl9xQ6E60TtJHSfbOxPTgBuxSecoDHMWveAOD+EiOFLwPfqSU8RLX5QweneVwIgQm9v9wU/veSHx//HEOBh3/leZUjn0DcAHpd4NxpavHqH
SN6IwXOGIMloRpjqio7jouSEBdYBk0DY0mI4jo4hpfG2qn/GXDLsy+zVfQQySPrFaRHW/C/L2lza09s96ZeXX/K+B8r+S0Sw7wWf9jxvHEetFRJITQlzid38rLA84lQv
ZCowhM3NCdFwemwaRRUK0IiFyUBCRhwwjvIjLfpF4YLxLJCsETSbhBvRrZQSnu/D94OI+NS4w39dyjARxgaefyCAa/7tOeAWM3es7LP7R6spWJsALkmrUFKA47RLUiSQ
Ka7An9+iwWy1yhT5MIzWq8zZfuAhVOKk8XqGEcOqaDXYIcd9StR9Y4uZ40AELwrv5zTGhrmg19v7pAtKxnMLHm994r4Hq//WHHC3WXv+IJASElLD+lIyTMzdtLccdD+w
tkGTxZzx2GYBaVaCFeI8jrRmgSQJ6QZRoEPaOThOu4SB0rFhAWet7jRFU0oJPwi0+oH/Ss6XYQ5Tt73gifsePPLfmgD33H3OmGeefxHLlq2A53uWfmm6hkcKQea2XrVq
fUbHIbb2Mtd0OYmwbayILVhTdMvU2mBOWy0xQ7NqWeGAKmHKQEKy1KxqfjeiiDb6h1y9mQBACIwYNhTbbL3t2F//u+uAs3edKT+4x25YvGw5HOFkxCcRASJqakhqsGcc
AJCKbko8I0oTbMWPHEbthx3b1T7ECTcVaUt5kxNTpoGJ+hIp27aM0gw6rXWt+hED1wzDUmYllVM1TBQ1QWm6o1dfiES4jLBGVa9l1lJI1coO6mRhMGZM2RwvLHiJ8zXp
fw8CLFerVQIE+vpqiTKvdkaPscCwiWHU2lXpOxzSklAaXStd10Ucrh8SqQCBBcEBAclvSpemKJst07EdnJyDEwKNAg6iUTNTFPESZwwTSIb7JD3iWBffkkI2msA1ERUE
jKS8h9qdCZChuy/KO5FB6JdOuxohifbWOkAZ2xBvizFLTqs8hBNVoK9WRa1aK0WYYP3fkgCJiLacMX3+Vw4+ZLlTLk9paxs0tFxuGemUSh2OEA7FjeTC1y7ViONEFCvf
k9avlILYMYRDRGkMqcpJEyM6Ij9KuWDc8BACelQOc0T0WXPbbILNoe9OqvpjButLSnioojv7CaaU6yVtvmREU6n8TrilZghFtxnvGOu4zExgDpiZA1mV0l/he/WuYZ0d
i278/Y1XCyFIyvcvY/T9trsEgKEAxgEYAaAtmnEiGouj7EfA/9matv+/jFsBMJVGc0CAsEGTH3G7XgBrACwHsB7vY6Mi+he9HCdahfJpOjoGlv4nRBiEqBJk/Pm+Lv8P
kvDhTXqzHXkAAAAASUVORK5CYILhb4B15U8AAA=="

  local -r VAR="$1"
  local FILE=
  local DIGESTALG=
  local DIGESTSUM=
  local LINE=
  local CHKSUM=
  local -i RETVAL=0

  {

    # Read metadata lines:
    #
    #   (1) [FILENAME]
    #   (2) [DIGESTALG]=[DIGESTSUM]
    #
    read -rs FILE
    IFS="=" read -rs DIGESTALG DIGESTSUM

    # Check for supplied path
    if (( ${2+1} )); then
      FILE="${2}/${FILE}"
    fi

    # Old openssl expects 64 chars per line
    while read -rs -n 64 LINE; do
      echo "$LINE"
    done | openssl enc -base64 -d | "$ZBIN" -cd > "$FILE"
    RETVAL=$?

  } < <(eval "echo \"\$${VAR}\"")

  # Check integrity of expanded file and report filename
  if ! (( RETVAL )); then
    IFS=" =" read -rs _ CHKSUM < <(openssl dgst -"$DIGESTALG" "$FILE")
    if [[ "$DIGESTSUM" == "$CHKSUM" ]]; then
      echo "$FILE"
    else
      echo "${FUNCNAME[0]}(): checksum of file \`${FILE}' is wrong" >&2
      RETVAL=1
    fi
  fi
  return $RETVAL
}

# Get username and if requested, the password if possible
get_credentials() {
  local -r DEFAULTUSR="$1"
  local -r HOST="$2"
  local TMPDATADIR=
  local GUIICON=
  local GUIIMAGE=
  local DIALOGBIN=
  local USERNAMELOG=
  local USERPASSLOG=
  local USR=
  local -i RETVAL=0

  if (( USEGUI )); then

    # Clean up on exit
    trap "rm -rf \"\$TMPDATADIR\"" EXIT

    # Setup the temporary data directory
    TMPDATADIR="$(mktemp -d -q -t "${BASENAME}-XXXXXXXXXXXX")" ||
      exit_message 1 "unable to create temporary directory, exiting"

    # Expand GUI data files
    GUIICON="$(expand_file "ICONDATA" "$TMPDATADIR")" ||
      exit_message 1 "unable to expand embedded file \`ICONDATA', exiting"
    GUIIMAGE="$(expand_file "IMAGEDATA" "$TMPDATADIR")" ||
      exit_message 1 "unable to expand embedded file \`IMAGEDATA', exiting"

    # Determine a usable `dialog' command
    if read -rs DIALOGBIN < <(command -v ${DIALOGBINS//"$SEPARATOR"/ }); then
      USERNAMELOG="acquiring username for [@${HOST}] via [${DIALOGBIN##*/}] GUI"
      USERPASSLOG="acquiring username/password for [@${HOST}] via [${DIALOGBIN##*/}] GUI"
      case "$DIALOGBIN" in

        */yad)
          if (( ASKPASS )); then
            local PASS=
            syslog "$USERPASSLOG"
            PASS="$("$DIALOGBIN" \
              --title="NX-Launch $VERSION" \
              --window-icon="$GUIICON" \
              --borders="15" \
              --fixed \
              --on-top \
              --skip-taskbar \
              --selectable-labels \
              --image="$GUIIMAGE" \
              --text-align="center" \
              --text="NX-Launch ${VERSION}\n-\nSystem Login" \
              --form \
              --align="right" \
              --separator="$SEPARATOR" \
              --field="Username @${HOST}:" "$DEFAULTUSR" \
              --field="Password:":H)"
            RETVAL=$?
            if ! (( RETVAL )); then
              IFS="$SEPARATOR" read -rs USR PASS <<< "$PASS"
              RETVAL=$?
            else
              # Try to get rid of the password in memory
              PASS="${RANDOM}${RANDOM}${RANDOM}${RANDOM}${RANDOM}${RANDOM}${RANDOM}"
              unset -v PASS
            fi
          else
            syslog "$USERNAMELOG"
            USR="$("$DIALOGBIN" \
              --title="NX-Launch $VERSION" \
              --window-icon="$GUIICON" \
              --borders="15" \
              --fixed \
              --on-top \
              --skip-taskbar \
              --selectable-labels \
              --entry \
              --text="Username @${HOST}:" \
              --entry-text="$DEFAULTUSR")"
            RETVAL=$?
          fi
          ;;

        */Xdialog)
          if (( ASKPASS )); then
            local PASS=
            syslog "$USERPASSLOG"
            PASS="$("$DIALOGBIN" \
              --title="NX-Launch $VERSION" \
              --icon="$GUIIMAGE" \
              --stdout \
              --separator="$SEPARATOR" \
              --password="2" \
              --2inputsbox="NX-Launch ${VERSION}\n-\nSystem Login" 0 0 \
              "Username @${HOST}:" "$DEFAULTUSR" "Password:" "")"
            RETVAL=$?
            if ! (( RETVAL )); then
              IFS="$SEPARATOR" read -rs USR PASS <<< "$PASS"
              RETVAL=$?
            else
              # Try to get rid of the password in memory
              PASS="${RANDOM}${RANDOM}${RANDOM}${RANDOM}${RANDOM}${RANDOM}${RANDOM}"
              unset -v PASS
            fi
          else
            syslog "$USERNAMELOG"
            USR="$("$DIALOGBIN" \
              --title="NX-Launch $VERSION" \
              --icon="$GUIIMAGE" \
              --stdout \
              --backtitle="NX-Launch ${VERSION}  -  System Login" \
              --inputbox="Username @${HOST}:" 0 0 "$DEFAULTUSR")"
            RETVAL=$?
          fi
          ;;

        */zenity)
          if (( ASKPASS )); then
            local PASS=
            syslog "$USERPASSLOG"
            PASS="$("$DIALOGBIN" \
              --title="Username @${HOST}:" \
              --password \
              --username)"
            RETVAL=$?
            if ! (( RETVAL )); then
              IFS="$SEPARATOR" read -rs USR PASS <<< "$PASS"
              RETVAL=$?
            else
              # Try to get rid of the password in memory
              PASS="${RANDOM}${RANDOM}${RANDOM}${RANDOM}${RANDOM}${RANDOM}${RANDOM}"
              unset -v PASS
            fi
          else
            syslog "$USERNAMELOG"
            USR="$("$DIALOGBIN" \
              --title="NX-Launch $VERSION" \
              --window-icon="$GUIICON" \
              --width="225" \
              --height="125" \
              --entry \
              --text="Username @${HOST}:" \
              --entry-text="$DEFAULTUSR")"
            RETVAL=$?
          fi
          ;;

        */kdialog)
          syslog "$USERNAMELOG"
          USR="$("$DIALOGBIN" \
            --caption="NX-Launch $VERSION" \
            --icon="preferences-desktop-user-password" \
            --inputbox="Username @${HOST}:" "$DEFAULTUSR")"
          RETVAL=$?
          ;;

        *)
          # We don't treat this as an error, just fall back to the terminal
          DIALOGBIN=

      esac
    fi

    # Remove temporary data directory
    rm -rf "$TMPDATADIR"

    # Nothing to clean up on exit anymore
    trap - EXIT

    if ! [[ "$DIALOGBIN" ]]; then
      syslog "unable to find a usable \`dialog' program, disabling GUI"
      syslog "(check installed packages like \`${DIALOGBINS%%"$SEPARATOR"*}' to fix this issue)"
    fi
  fi

  if ! [[ "$DIALOGBIN" ]]; then
    if (( BASH_VERSINFO[0] < 4 )); then
      # BASH < 4 `read' doesn't support default values with `-i'
      # shellcheck disable=SC2162 # we don't want raw mode here
      read -ep "Username @${HOST} [${DEFAULTUSR}]: " USR
      RETVAL=$?
      USR="${USR:-${DEFAULTUSR}}"
    else
      # shellcheck disable=SC2162 # we don't want raw mode here
      read -ep "Username @${HOST}: " -i "$DEFAULTUSR" USR
      RETVAL=$?
    fi
  fi

  # Strip bad spaces
  USR="$(sanitize_list " " "$USR")"

  # Username must not be empty
  if ! [[ "$USR" ]]; then
    RETVAL=1
  fi

  # If all is well so far, assign username and handle password if we got it
  if (( RETVAL )); then
    syslog "username is not set, exiting"
  else
    DSTUSER="$USR"
    # Handle the password in case we got it
    if (( ${PASS+1} )); then
      # At this stage `$SSH_ASKPASS' is set to the original `$SSH_ASKPASS' binary
      if handle_password "${DSTUSER}@${DSTHOST}"; then
        # There was no error, show that we have handled the password
        RETVAL=2
      else
        # There was an error handling password
        RETVAL=1
      fi
    fi
  fi
  if (( ${PASS+1} )); then
    # Try to get rid of the password in memory
    PASS="${RANDOM}${RANDOM}${RANDOM}${RANDOM}${RANDOM}${RANDOM}${RANDOM}"
    unset -v PASS
  fi
  return $RETVAL
}

# Get and submit password
handle_password() {
  local -r DST="$1"
  local -i RETVAL=0

  # Check if `get_credentials()' already provided us a password
  if ! (( ${PASS+1} )); then
    local PASS=
    # Don't report unusable `$SSH_ASKPASS' as it was already reported before
    if (( USEGUI )) && [[ "$SSH_ASKPASS" ]]; then
      syslog "acquiring password for [${DST}] via GUI"
      PASS="$("$SSH_ASKPASS" "${DST}'s password: " 2> /dev/null)"
      RETVAL=$?
    else
      IFS= read -rsp "${DST}'s password: " PASS
      RETVAL=$?
      echo >&2
    fi
  fi

  # Don't submit password if anything went wrong
  if ! (( RETVAL )); then
    # Generate a random encryption password
    CRYPTRAND="$(uuidgen 2>> "$NXPROXYLOG")" ||
      exit_message 1 "unable to generate random string, exiting"

    # Encrypt the password which will be submitted via FIFO
    CRYPTPASS="$(CRYPTRAND="$CRYPTRAND" openssl enc -aes-256-cbc -salt -a -A -pass env:CRYPTRAND 2>> "$NXPROXYLOG" <<< "$PASS")" ||
      exit_message 1 "unable to encrypt password, exiting"

    # Try to get rid of the password in memory
    PASS="${RANDOM}${RANDOM}${RANDOM}${RANDOM}${RANDOM}${RANDOM}${RANDOM}"
    unset -v PASS

    # Activate built in SSH authentication mechanism by setting `$SSH_ASKPASS' to ourself
    syslog "establishing built in SSH authentication mechanism"
    export SSH_ASKPASS="${LIBDIRABS}/${BASENAME}"
    command -v "$SSH_ASKPASS" > /dev/null ||
      exit_message 1 "unable to execute ${SSH_ASKPASS}, exiting"

    # Force SSH to run in new session so that it will execute `$SSH_ASKPASS'
    SSHSETSID=1
  else
    syslog "password is not set, exiting"
  fi

  return $RETVAL
}

# Run `$SSH_ASKPASS' password authentication mechanism called from SSH
run_ssh_askpass() {
  local -ri AUTHPID=$$
  local -i CHLDPID=
  # SSH option `LoginGraceTime' defaults to 120 seconds
  local -i COUNT=60
  local PASS=
  local -i RETVAL=0

  # Clean up on exit
  trap "rm -rf \"\${SSH_AUTH_FIFO%/*}\"" EXIT

  {
    # Switch BASH into restricted more for increased security
    set -r

    # Unable to do anything with empty `$SSH_AUTH_FIFO' variable
    if ! [[ "$SSH_AUTH_FIFO" ]]; then
      return 1
    fi

    # Start a monitoring background task
    {
      for (( COUNT = COUNT * SHORTPERSEC; COUNT; COUNT-- )); do
        # Go on if the communication FIFO or the parent process have vanished
        if ! [[ -p "$SSH_AUTH_FIFO" ]] || ctrl_proc $AUTHPID; then
          break
        fi
        sleep "$SHORTDELAY"
      done

      # Terminate the parent process which should already have exited. This
      # function will report harmless errors in BASH restricted more.
      ctrl_proc $AUTHPID -3

      # Don't return the exit status code of the command above
      return 0
    } &
    CHLDPID=$!

    # Terminate child on `SIGTERM'
    trap "kill \$CHLDPID" SIGTERM

    # Read encrypted password from FIFO
    # shellcheck disable=SC1007 # shellcheck doesn't understand this
    ILS= read -rs PASS < "$SSH_AUTH_FIFO"
    (( RETVAL+=$? ))
    # Decrypt password and submit it to SSH
    SSH_AUTH_RAND="$SSH_AUTH_RAND" \
      openssl enc -aes-256-cbc -salt -a -A -pass env:SSH_AUTH_RAND -d <<< "$PASS"
    (( RETVAL+=$? ))

    # Wait for child to terminate to avoid a zombie process
    wait $CHLDPID

    return $RETVAL
  } 2>> "$NXPROXYLOG"
}

# Submit the password to `run_ssh_askpass()' called as `$SSH_ASKPASS' from SSH
submit_cryptpass() {
  local -i CHLDPID=
  # SSH option `LoginGraceTime' defaults to 120 seconds
  local -i COUNT=60

  # Submit the password via FIFO in background task
  echo "$CRYPTPASS" > "$CRYPTFIFO" &
  CHLDPID=$!

  for (( COUNT = COUNT * SHORTPERSEC; COUNT; COUNT-- )); do
    # Go on if the communication FIFO or the child process have vanished
    if ! [[ -p "$CRYPTFIFO" ]] || ctrl_proc $CHLDPID; then
      break
    fi
    sleep "$SHORTDELAY"
  done

  # Terminate the child process which should already have exited
  ctrl_proc $CHLDPID -3

  # Remove the FIFO in case it still exists
  if [[ -p "$CRYPTFIFO" ]]; then
    rm -rf "${CRYPTFIFO%/*}"
  fi

  # Don't return the exit status code of the previous command
  return 0
}

# Shutdown command/NX agent and exit
shutdown_agent_and_exit() {
  local -i EXITVAL=$RETVAL

  # Don't re-exec ourself
  # shellcheck disable=SC2086 # doesn't apply to integer variables or lists
  trap - $EXITSIGNALS

  if (( CMDPID )) && ! ctrl_proc $CMDPID; then
    syslog "shutting down command, pid [${CMDPID}]"
    if ctrl_proc $CMDPID -3; then
      # The process has already terminated, we only fetch its exit status code
      wait $CMDPID 2> /dev/null
      RETVAL=$?
      syslog "command has terminated, pid [${CMDPID}] status [${RETVAL}]"
      CMDPID=0
    else
      syslog "command did not terminate, pid [${CMDPID}]"
      RETVAL=1
    fi
    if (( RETVAL > EXITVAL )); then
      EXITVAL=$RETVAL
    fi
  fi

  if (( AGENTPID )) && ! ctrl_proc $AGENTPID; then
    syslog "shutting down NX agent, pid [${AGENTPID}]"
    if ctrl_proc $AGENTPID -3; then
      # The process has already terminated, we only fetch its exit status code
      wait $AGENTPID 2> /dev/null
      RETVAL=$?
      syslog "NX agent has terminated, pid [${AGENTPID}] status [${RETVAL}]"
      AGENTPID=0
    else
      syslog "NX agent did not terminate, pid [${AGENTPID}]"
      RETVAL=1
    fi
    if (( RETVAL > EXITVAL )); then
      EXITVAL=$RETVAL
    fi
  fi

  remove_cookie $NXDISPLAY >> "$NXAGENTLOG" 2>&1
  # Run post command
  if [[ "$NXPOSTCMD" ]]; then
    run_system "$NXPOSTCMD" "running post command"
  fi

  syslog "exiting with exit status code [${EXITVAL}]"
  exit $EXITVAL
}

# Run a remote command in client/server mode
run_rexec() {
  local REVPORT=
  REVPORT="${DISPLAY##*:}"
  REVPORT=${REVPORT%%.*}
  REVPORT=$(( REXECBASEPORT + REVPORT ))
  socat -d0 -t "$REXECTIMEOUT" TCP-CONNECT:localhost:"${REVPORT}" STDIO <<< "$REMCMD"
}

# Handling NX agent for client/server and local mode
run_agent() {
  local ACCEPTHOST=
  local AGENTOPT=
  local -i TIMESTAMP=
  local DISPOPTS=
  local -i NXDISPLAY=
  local XCOOKIE=
  local PROXYHOST=
  local -i CMDPID=0
  local -i AGENTPID=0
  local -i RETVAL=0

  # Assign signal handlers to clean up and exit
  # shellcheck disable=SC2086 # doesn't apply to integer variables or lists
  trap "shutdown_agent_and_exit" $EXITSIGNALS

  # Run NX agent standalone or with NX proxy
  if ! (( NXPROXYDISPLAY )); then
    # Tag logs
    LOGID="${BASENAME}[${$}]: local"

    log_version_config

    # NX agent parameters, we use `$DISPLAY' from environment
    DISPOPTS="$DISPLAY"
    NXDISPLAY=$(get_free_agent_display 2>> "$NXAGENTLOG") || return 1
    XCOOKIE="$(get_cookie 2>> "$NXAGENTLOG")" || return 1
  else
    # Tag logs
    LOGID="${BASENAME}[${$}]: server"

    log_version_config

    # NX agent/proxy options
    NXDISPLAY=$NXPROXYDISPLAY
    DISPOPTS="nx/nx,${NXPROXYOPTS}"
    DISPOPTS="${DISPOPTS}${NXADDOPTS:+,${NXADDOPTS}}"
    if (( NXUNENCRYPTED )); then
      ACCEPTHOST="${SSH_CONNECTION%% *}"
    else
      ACCEPTHOST="$LOCALHOST"
    fi
    DISPOPTS="${DISPOPTS},accept=${ACCEPTHOST}"
    DISPOPTS="${DISPOPTS}${NXLINK:+,link=${NXLINK}}"
    DISPOPTS="${DISPOPTS}${NXPACK:+,pack=${NXPACK}}"
    DISPOPTS="${DISPOPTS}${NXLIMIT:+,limit=${NXLIMIT}}"
    DISPOPTS="${DISPOPTS}:${NXDISPLAY}"

    # Sanitize composed NX options string
    DISPOPTS="$(sanitize_list "," "$DISPOPTS")"

    # Assign options from remote instance
    XCOOKIE="$SHIPXCOOKIE"
    PROXYHOST="$SHIPHOSTNAME"

    # Useful information to identify the connection
    syslog "client host [${PROXYHOST}]"
    if [[ "$NXHOSTID" ]]; then
      syslog "client hostid [${NXHOSTID}]"
    fi
    syslog "connection [${SSH_CONNECTION}]"

    # Check that requested display number is free on local host
    syslog "client requesting display [${NXDISPLAY}]"
    if ! check_agent_display_free $NXDISPLAY; then
      send_notification "display $NXDISPLAY" "BUSY"
      return 1
    else
      send_notification "display $NXDISPLAY" "FREE"
    fi
  fi

  # Do we need to listen for incoming TCP connections on NX agent
  if ! (( NXLISTENTCP )); then
    NXAGENTOPTS="$NXAGENTOPTS -nolisten tcp"
  fi

  # NX agent window mode, rootless or with geometry/default size
  if (( NXROOTLESS )); then
    NXAGENTOPTS="$NXAGENTOPTS -R"
    if [[ "$NXGEO" ]]; then
      syslog "ignoring NX agent geometry options [${NXGEO}] in rootless mode"
    fi
  else
    NXAGENTOPTS="${NXAGENTOPTS}${NXGEO:+ -geometry $NXGEO}"
  fi

  # NX agent keyboard type, detect or use `clone' if undefined
  if [[ "$NXKBTYPE" == "detect" ]]; then
    NXKBTYPE="$(get_local_kbtype)"
  else
    NXKBTYPE="${NXKBTYPE:-clone}"
  fi
  # Don't set the keyboard at all if keyboard type is still undefined
  NXAGENTOPTS="${NXAGENTOPTS}${NXKBTYPE:+ -keyboard $NXKBTYPE}"

  # XDMCP or execute command
  if ! [[ "$NXCMD" ]]; then
    NXAGENTOPTS="$NXAGENTOPTS -query $NXQUERY"
    # We don't show `$LOCALHOST' here simply because it could be confusing
    # for the user as we connect to `$LOCALHOST' on the _remote_ machine.
    if [[ "$NXQUERY" == "$LOCALHOST" ]]; then
      NXTITLE="${NXTITLE:-NX-Launch $VERSION - display:${NXDISPLAY} host:${HOSTNAME} XDMCP}"
    else
      NXTITLE="${NXTITLE:-NX-Launch $VERSION - display:${NXDISPLAY} host:${HOSTNAME} XDMCP:${NXQUERY}}"
    fi
  else
    NXTITLE="${NXTITLE:-NX-Launch $VERSION - display:${NXDISPLAY} host:${HOSTNAME} user:${USER} command:${NXCMD}}"
  fi

  # NX agent common options
  if (( NXAGENTAUTODPI )); then
    AGENTOPT="-autodpi"
    if check_agent_option "$AGENTOPT"; then
      NXAGENTOPTS="$NXAGENTOPTS $AGENTOPT"
    else
      syslog "NX agent omitting unsupported option [${AGENTOPT}]"
    fi
  fi
  if ! (( NXAGENTMAGICPIXEL )); then
    AGENTOPT="-nomagicpixel"
    if check_agent_option "$AGENTOPT"; then
      NXAGENTOPTS="$NXAGENTOPTS $AGENTOPT"
    else
      syslog "NX agent omitting unsupported option [${AGENTOPT}]"
    fi
  fi
  if ! (( NXAGENTRENDER )); then
    NXAGENTOPTS="$NXAGENTOPTS -norender"
  fi
  if ! (( NXAGENTSHMEM )); then
    NXAGENTOPTS="$NXAGENTOPTS -noshmem"
  fi
  if ! (( NXAGENTSHPIX )); then
    NXAGENTOPTS="$NXAGENTOPTS -noshpix"
  fi

  # Run pre command
  if [[ "$NXPRECMD" ]]; then
    run_system "$NXPRECMD" "running pre command"
  fi

  # Check hostname resolution and roughly measure round-trip time (RTT) to remote host
  if ! [[ "$NXCMD" ]] && [[ "$NXQUERY" != "$LOCALHOST" ]]; then
    check_remote_host "$NXQUERY" >> "$NXAGENTLOG" || return 1
  fi

  # Add X authorization cookie (MIT-MAGIC-COOKIE-1) entries for NX agent
  add_cookie $NXDISPLAY "$XCOOKIE" >> "$NXAGENTLOG" 2>&1

  # Start NX agent
  syslog "starting [$(get_agent_version)] for display [${NXDISPLAY}]"
  # shellcheck disable=SC2086 # doesn't apply to integer variables or lists
  exec "$NXAGENTBIN" $NXAGENTOPTS -display $DISPOPTS -name "$NXTITLE" :${NXDISPLAY} >> "$NXAGENTLOG" 2>&1 &
  AGENTPID=$!

  # Wait for NX agent to accept connections and tell NX proxy about it
  if check_agent_listen $AGENTPID $NXDISPLAY; then
    syslog "NX agent has started, pid [${AGENTPID}]"
    if (( NXPROXYDISPLAY )); then
      send_notification "agent $NXDISPLAY" "READY"
    fi
  else
    syslog "NX agent not accepting connections, exiting"
    if (( NXPROXYDISPLAY )); then
      send_notification "agent $NXDISPLAY" "ERROR"
    fi
    RETVAL=1
    shutdown_agent_and_exit
  fi

  # Start command and/or wait for XDMCP/command to terminate
  export DISPLAY=":${NXDISPLAY}.0"
  if ! [[ "$NXCMD" ]]; then
    wait $AGENTPID 2> /dev/null
    RETVAL=$?
    syslog "NX agent has terminated, pid [${AGENTPID}] status [${RETVAL}]"
    AGENTPID=0
  else
    if (( NXPROXYDISPLAY )); then
      syslog "waiting for connection to NX proxy"
    else
      syslog "waiting for display [${DISPLAY}] on NX agent to become ready"
    fi
    TIMESTAMP=$(( $(date "+%s") + AGENTSTARTUPTIMEOUT ))
    while (( $(date "+%s") < TIMESTAMP )); do
      # Check that X via NX proxy is really working
      if xwininfo -root > /dev/null 2>&1; then
        if (( NXPROXYDISPLAY )); then
          syslog "connection to NX proxy established"
        else
          syslog "display [${DISPLAY}] on NX agent is ready"
        fi
        syslog "running command [${NXCMD}]"
        eval exec "$NXCMD" &
        CMDPID=$!
        syslog "command has started, pid [${CMDPID}]"
        wait $CMDPID 2> /dev/null
        RETVAL=$?
        syslog "command has terminated, pid [${CMDPID}] status [${RETVAL}]"
        CMDPID=0
        break
      else
        syslog "display [${DISPLAY}] on NX agent is not ready"
        sleep 1
      fi
    done
  fi

  # Clean up and exit
  shutdown_agent_and_exit
}

# Handling NX proxy for client/server mode
run_proxy() {
  local DSTUSER=
  local DSTHOST=
  local DSTRTT=
  local PROXYPIDSTR=
  local PROXYVER=
  local SSHVER=
  local OSVER=
  local CORE=
  local CPU=
  local _NET_DESKTOP_GEOMETRY=
  local -i NXDISPLAY=
  local XCOOKIE=
  local POPTS=
  local PROXYOUT=
  local DISPSTATUS=
  local AGENTSTATUS=
  local TRACEFILE=
  local SSHCMD="$SSHBIN"
  local SSHOPTS=
  local SSHERR=
  local CRYPTRAND=
  local CRYPTFIFO=
  local CRYPTPASS=
  local SSH_AUTH_RAND=
  local SSH_AUTH_FIFO=
  local -i HANDLEDPASS=0
  local -i SSHSETSID=0
  local -i SSHPORT=
  local -i REVSHELLPORT=
  local -i REVEXECPORT=
  local -i TIMESTAMP=
  local -i COUNT=1
  local -ri MAXTRY=7
  local -i TRY=
  local -i SOCATPID=
  local -i RETVAL=0

  # Tag logs
  LOGID="${BASENAME}[${$}]: client"

  log_version_config

  # Split `username' @ `hostname:port' of `location' string
  if ! [[ "${NXLOCATION%%*@*}" ]]; then
    DSTHOST="${NXLOCATION##*@}"
    DSTUSER="${NXLOCATION%%@*}"
  else
    DSTHOST="$NXLOCATION"
  fi

  # Split `hostname' : `port' of `hostname:port' string
  if ! [[ "${DSTHOST%%*:*}" ]]; then
    RSSHPORT="${DSTHOST##*:}"
    DSTHOST="${DSTHOST%%:*}"
  fi

  # Provide simple GUI as requested and enforce SSH to execute
  # `ssh-askpass' if a password is required.
  if (( USEGUI )); then
    # The `$SSH_ASKPASS' variable is most likely already set
    if ! (( ${SSH_ASKPASS+1} )); then
      # Determine a usable `ssh-askpass' command
      read -rs SSH_ASKPASS < <(command -v ${ASKPASSBINS//"$SEPARATOR"/ })
    fi
    if (( ${SSH_ASKPASS+1} )) && command -v "$SSH_ASKPASS" > /dev/null; then
      export SSH_ASKPASS
      # Force SSH to run in new session so that it will execute `$SSH_ASKPASS'
      SSHSETSID=1
    else
      syslog "unable to find a usable \`ssh-askpass' program, disabling SSH askpass"
      syslog "(check installed packages like \`openssh-askpass' to fix this issue)"
      unset -v SSH_ASKPASS
    fi
  fi

  # Get username and if requested, the password if possible
  if (( ASKUSER )); then
    if [[ "$DSTUSER" ]]; then
      syslog "discarding implicitly declared username [${DSTUSER}]"
    fi
    get_credentials "$DSTUSER" "$DSTHOST"
    RETVAL=$?
    if (( RETVAL == 1 )); then
      # We had an error which is final
      return 1
    elif (( RETVAL == 2 )); then
      # We handled the password in `get_credentials()', don't do it again
      HANDLEDPASS=1
      RETVAL=0
    fi
  elif ! [[ "$DSTUSER" ]]; then
    # Fall back to our own username
    DSTUSER="$USER"
  fi

  # Get the password as requested and don't execute the usual `ssh-askpass' from SSH
  if (( ASKPASS )) && ! (( HANDLEDPASS )); then
    # At this stage `$SSH_ASKPASS' is set to the original `$SSH_ASKPASS' binary
    handle_password "${DSTUSER}@${DSTHOST}" || return 1
  fi

  # Run SSH in a new session if required
  if (( SSHSETSID )); then
    SSHCMD="setsid $SSHCMD"
  fi

  # Check hostname resolution and roughly measure round-trip time (RTT) to remote host
  DSTRTT="$(check_remote_host "$DSTHOST")" || return 1

  # Collect some useful info to show in the process title
  # NX proxy version
  PROXYVER="$(get_proxy_version)"
  PROXYVER="${PROXYVER##* }"
  # Shortened SSH version
  SSHVER="$($SSHBIN -V 2>&1)"
  SSHVER="${SSHVER%,*}"
  SSHVER="${SSHVER#Open}"
  # OS name and version
  OSVER="$(uname -o), $(uname -r), $(uname -m)"
  # CPU core(s) version(s) on Linux...
  while IFS=" :" read -rs _ _ CORE; do
    if [[ "$CORE" == "-" ]]; then
      CORE="(CPU core(s) unknown)"
    fi
    CPU="${CPU}${CPU:+ + }${CORE}"
  done < <(LC_ALL="C" lscpu 2>> "$NXPROXYLOG" | match_lines "#Model name")
  if ! [[ "$CPU" ]]; then
    # ...or Linux (old)
    IFS=" :" read -rs _ _ CPU < <(match_lines "#model name" < /proc/cpuinfo 2>> "$NXPROXYLOG")
    if ! [[ "$CPU" ]]; then
      # ...or FreeBSD
      if ! CPU="$(sysctl -n hw.model 2>> "$NXPROXYLOG")"; then
        CPU="(CPU core(s) not detected)"
      fi
    fi
  fi
  CPU="$(sanitize_list " " "$CPU")"
  # Geometry of the X root window
  eval "$(LC_ALL="C" xprop -root -notype -f _NET_DESKTOP_GEOMETRY "32c" "=\"\$0x\$1\"\n" _NET_DESKTOP_GEOMETRY 2>> "$NXPROXYLOG")"

  # Repeat until we get a free remote display number
  while :; do

    # Get free display number
    NXDISPLAY=$(get_free_proxy_display 2>> "$NXPROXYLOG") || return 1
    # Get local X authorization cookie (MIT-MAGIC-COOKIE-1)
    XCOOKIE="$(get_cookie 2>> "$NXPROXYLOG")" || return 1

    # Define SSH and reverse tunnel ports
    SSHPORT=$(( NXBASEPORT + NXDISPLAY ))
    REVSHELLPORT=$(( REVBASEPORT + NXDISPLAY ))
    REVEXECPORT=$(( REXECBASEPORT + NXDISPLAY ))

    # Parameters for SSH and NX agent
    POPTS="--proxydisplay $NXDISPLAY"
    SSHOPTS="$SSHX11FORWARD"

    # Run the remote shell helper early
    if (( ENABLEREXEC )); then
      SSHOPTS="${SSHOPTS} -R ${REVEXECPORT}:${LOCALHOST}:${REVEXECPORT}"
      exec socat -d0 TCP-LISTEN:${REVEXECPORT},fork,reuseaddr EXEC:"${BASHBIN}",nofork,stderr &
      SOCATPID=$!
      sleep "$SHORTDELAY"
    fi

    SSHOPTS="${SSHOPTS} -R ${REVSHELLPORT}:${LOCALHOST}:${LSHELLPORT}"
    SSHOPTS="${SSHOPTS}${NXDEBUG:+${SSHOPTS:+ }-${NXDEBUG}}"
    SSHOPTS="${SSHOPTS}${RSSHPORT:+${SSHOPTS:+ }-p $RSSHPORT}"
    SSHOPTS="${SSHOPTS}${SSHCIPHERS:+${SSHOPTS:+ }-c $SSHCIPHERS}"

    if ! [[ "$LOGSTDERR" ]]; then
      POPTS="${POPTS}${POPTS:+ }--quiet"
    fi
    if (( NXUNENCRYPTED )); then
      NXAGENTHOST="$DSTHOST"
      POPTS="${POPTS}${POPTS:+ }--unencrypted"
    else
      NXAGENTHOST="$LOCALHOST"
      SSHOPTS="${SSHOPTS}${SSHOPTS:+ }-L ${SSHPORT}:${LOCALHOST}:${SSHPORT}"
    fi
    if [[ "$CRYPTPASS" ]]; then
      # Prompt only once for password with built in SSH authentication mechanism
      SSHOPTS="${SSHOPTS}${SSHOPTS:+ }-o NumberOfPasswordPrompts=1"
      # Don't prompt for adding host keys with built in SSH authentication mechanism
      SSHOPTS="${SSHOPTS}${SSHOPTS:+ }-o StrictHostKeyChecking=no -o CheckHostIP=no"
    fi
    if (( NXROOTLESS )); then
      POPTS="${POPTS}${POPTS:+ }--rootless"
    fi
    if [[ "$NXAGENTLOG" != "$DEFAULTLOG" ]]; then
      POPTS="${POPTS}${POPTS:+ }--nxagentlog \"${NXAGENTLOG}\""
    fi
    POPTS="${POPTS}${NXHOSTID:+${POPTS:+ }--hostid \"${NXHOSTID}\"}"
    POPTS="${POPTS}${NXGEO:+${POPTS:+ }--geometry \"${NXGEO}\"}"
    POPTS="${POPTS}${NXKBTYPE:+${POPTS:+ }--keyboard \"${NXKBTYPE}\"}"
    POPTS="${POPTS}${NXTITLE:+${POPTS:+ }--title \"${NXTITLE}\"}"
    POPTS="${POPTS}${NXLINK:+${POPTS:+ }--compression \"${NXLINK}\"}"
    POPTS="${POPTS}${NXPACK:+${POPTS:+ }--imagepack \"${NXPACK}\"}"
    POPTS="${POPTS}${NXLIMIT:+${POPTS:+ }--bandwidth \"${NXLIMIT}\"}"
    POPTS="${POPTS}${NXADDOPTS:+${POPTS:+ }--nxoptions \"${NXADDOPTS}\"}"
    if ! [[ "$NXCMD" ]]; then
      POPTS="${POPTS}${NXQUERY:+${POPTS:+ }--query \"${NXQUERY}\"}"
    fi
    # Commands in `$NXCMD' must be the last NX agent parameter
    POPTS="${POPTS}${NXCMD:+${POPTS:+ }-- $NXCMD}"

    # Run pre command
    if [[ "$NXPRECMD" ]]; then
      run_system "$NXPRECMD" "running pre command"
    fi

    # Ship code and X authorization cookie (MIT-MAGIC-COOKIE-1) to remote
    # host and start remote NX agent, then start local NX proxy and intercept
    # STDERR of `$SSHCMD' to catch remote port forwarding failures.
    # This is because not all SSH have the `ExitOnForwardFailure' feature
    # to exit on failure. We shut down the running SSH in case for security
    # reasons even if the NX session would be usable.
    # To test this, set `$REVSHELLPORT' in the `$SSHCMD' call to an already
    # used port.
    syslog "shipping vital components via [${DSTUSER}@${DSTHOST}${RSSHPORT:+:$RSSHPORT}]"
    syslog "starting $BASENAME on host [${DSTHOST}] port [${SSHPORT}] try [${COUNT}/${MAXTRY}]"

    # Activate built in SSH authentication mechanism if enabled
    if [[ "$CRYPTPASS" ]]; then
      # Setup the communication FIFO
      CRYPTFIFO="$(mktemp -d -q -t "${BASENAME}-XXXXXXXXXXXX" 2>> "$NXPROXYLOG")/${BASENAME}.${$}" ||
        exit_message 1 "unable to create temporary directory for FIFO, exiting"
      mkfifo -m 0600 "$CRYPTFIFO" 2>> "$NXPROXYLOG" ||
        exit_message 1 "unable to create FIFO ${CRYPTFIFO}, exiting"

      # Submit the encrypted password via FIFO
      submit_cryptpass 2>> "$NXPROXYLOG" &
    fi

    # Prepare shipping and launch via SSH tunnel
    # shellcheck disable=SC2086 # doesn't apply to integer variables or lists
    {

      # Generating the `nxlaunch' program for the remote host: START
      echo "# $BASENAME version $VERSION"
      echo "# Generated on \`${HOSTNAME}' at \`$(date -R)'"
      echo
      echo "{"
      echo

      # Clear locale related vars which SSH may have received
      if (( NOLOCALE )); then
        # shellcheck disable=SC1083,SC2116 # shellcheck doesn't understand this
        echo "  unset -v $(echo LANG LANGUAGE \${!LC_@} XMODIFIERS)"
        echo
      fi

      # Do what `bash --login' would do
      echo "  if [[ -r /etc/profile ]]; then"
      echo "    .      /etc/profile"
      echo "  fi"
      echo "  if [[ -r ~/.bash_profile ]]; then"
      echo "    .      ~/.bash_profile"
      echo "  elif [[ -r ~/.bash_login ]]; then"
      echo "    .        ~/.bash_login"
      echo "  elif [[ -r ~/.profile ]]; then"
      echo "    .        ~/.profile"
      echo "  fi"
      echo

      # Ship some needed vars
      echo "  declare -r BASENAME=\"${BASENAME}\""
      echo "  declare -r LIBDIR=\"${LIBDIR}\""
      echo "  declare -r SHIPXCOOKIE=\"${XCOOKIE}\""
      echo "  declare -r SHIPHOSTNAME=\"${HOSTNAME}\""
      echo "  declare -i PROCTITLE="
      echo

      # Ship applied configuration data
      echo "#${CUTHERE} Configuration data applied"
      if [[ "$CONFIGDATA" ]]; then
        echo "$CONFIGDATA"
      else
        echo "# (none)"
      fi
      echo "#${CUTHERE} Configuration data end"
      echo

      echo "} 2>> $NXAGENTLOG"
      echo

      # Ship main program
      echo "#${CUTHERE} Code from \`${LIBDIR}/${BASENAME}'"
      echo "$(< "${LIBDIR}/${BASENAME}")"
      echo "#${CUTHERE} Code end"
      echo
      # Generating the `nxlaunch' program for the remote host: END

    } 2>> "$NXPROXYLOG" |
      {
        if [[ "$NXDEBUG" ]]; then
          TRACEFILE="$NXPROXYLOG"
        fi
        copy_stdio_append "$TRACEFILE"
      } 2>> "$NXPROXYLOG" |
        "$ZBIN" -cv 2>> "$NXPROXYLOG" |
          SSH_AUTH_RAND="$CRYPTRAND" \
            SSH_AUTH_FIFO="$CRYPTFIFO" \
            $SSHCMD $SSHOPTS "${DSTUSER}@${DSTHOST}" \
            "exec -a \"NX: $BASENAME $VERSION" \
            "[Proxy (${PROXYVER}, ${_NET_DESKTOP_GEOMETRY}, ${CPU}, ${OSVER})," \
            "$SSHVER (${HOSTNAME}, \${SSH_CONNECTION%% *}, ${DSTRTT})]\"" \
            "$BASHBIN -s -- $POPTS < <($ZBIN -cd)" \
            2> >(
              while read -rs SSHERR; do
                if [[ "$SSHERR" != "${SSHERR/ort forwarding failed}" ]]; then
                  echo "FAIL"
                  pkill -P $$ -x "$SSHBIN"
                else
                  echo "$SSHERR" >&2
                fi
              done
            ) |
              {
                # We should get these lines of text on STDIN from NX agent or SSH error logger:
                #
                #   (1) [FREE|BUSY|FAIL] for display number free status
                #   (2) [READY|ERROR] for NX agent running status
                #
                syslog "waiting for remote NX agent startup notification"
                read -t $SSHTUNNELTIMEOUT -rs DISPSTATUS
                # Built in SSH authentication has finished at this stage,
                # remove the FIFO in case `$SSH_ASKPASS' failed to do so.
                if [[ -p "$CRYPTFIFO" ]]; then
                  rm -rf "${CRYPTFIFO%/*}"
                fi
                # Evaluate what we received before
                if [[ "$DISPSTATUS" == "FAIL" ]]; then
                  # Exit only from the `$SSHCMD' here
                  exit 98
                elif [[ "$DISPSTATUS" == "BUSY" ]]; then
                  # Exit only from the `$SSHCMD' here
                  exit 99
                fi
                read -t $AGENTREADYTIMEOUT -rs AGENTSTATUS
                if [[ "$AGENTSTATUS" == "READY" ]]; then
                  syslog "remote NX agent startup completed"
                  for (( TRY = 1; TRY <= MAXTRY; TRY++ )); do
                    syslog "starting [$(get_proxy_version)] for display [${NXDISPLAY}] try [${TRY}/${MAXTRY}]"
                    TIMESTAMP=$(( $(date "+%s") + PROXYSTARTUPTIMEOUT ))
                    RETVAL=0
                    PROXYPIDSTR="(unknown)"
                    while read -rs PROXYOUT; do
                      echo "$PROXYOUT" >> "$NXPROXYLOG"
                      if [[ "$PROXYOUT" != "${PROXYOUT/#Error:}" ]]; then
                        RETVAL=1
                      elif [[ "$PROXYOUT" != "${PROXYOUT/#Info: Proxy running in * mode with pid \'}" ]]; then
                        PROXYPIDSTR="${PROXYOUT#*\'}"
                        PROXYPIDSTR="${PROXYPIDSTR%\'*}"
                        syslog "NX proxy has started, pid [${PROXYPIDSTR}]"
                      fi
                    done < <(exec "$NXPROXYBIN" -S "nx/nx,${NXPROXYOPTS},connect=${NXAGENTHOST}:${NXDISPLAY}" 2>&1)
                    if (( RETVAL )); then
                      if (( $(date "+%s") < TIMESTAMP )); then
                        syslog "NX proxy startup error, pid [${PROXYPIDSTR}] status [${RETVAL}], retrying..."
                        sleep 1
                      else
                        syslog "NX proxy startup error, pid [${PROXYPIDSTR}] status [${RETVAL}]"
                        break
                      fi
                    else
                      syslog "NX proxy has terminated, pid [${PROXYPIDSTR}] status [${RETVAL}]"
                      break
                    fi
                  done
                else
                  # Exit only from the `$SSHCMD' here
                  exit 255
                fi
              }
              # End of `$SSHCMD' code running on the remote host

    # Get the exit status of `$SSHCMD'
    RETVAL=$?

    # Terminate the remote shell helper
    if (( ENABLEREXEC )); then
      ctrl_proc $SOCATPID -3
    fi

    # Evaluate the exit status of `$SSHCMD'
    case $RETVAL in
      98)
        syslog "unable to start SSH tunneling for display [${NXDISPLAY}], retrying"
        syslog "with another display number..."
        sleep 1
        ;;
      99)
        syslog "unable to start remote NX agent, display [${NXDISPLAY}] is already"
        syslog "used, retrying with another display number..."
        sleep 1
        ;;
      130)
        syslog "connection timed out or shutdown requested by user, exiting"
        break
        ;;
      255)
        syslog "unable to establish SSH connection to [${DSTUSER}@${DSTHOST}], exiting"
        break
        ;;
      *)
        break
    esac

    if (( COUNT++ >= MAXTRY )); then
      syslog "unable to get a free NX display number, exiting"
      break
    fi
  done

  # Run post command
  if [[ "$NXPOSTCMD" ]]; then
    run_system "$NXPOSTCMD" "running post command"
  fi

  syslog "exiting with exit status code [${RETVAL}]"
  return $RETVAL
}

#-----------------------------------------------------------------------------

# We were called as `$SSH_ASKPASS' from SSH
if (( ${SSH_AUTH_FIFO+1} )); then
  # SSH calls this with `$1' set to something like `user@host's password: '
  run_ssh_askpass
  exit
fi

# Search and apply configuration files
if ! (( NXPROXYDISPLAY )); then
  declare CONFIGITEM=

  # Get a list of valid configuration files
  CONFIGAPPLIED="$(
    while read -rs CONFIGITEM; do
      if [[ -r "$CONFIGITEM" ]] && [[ -s "$CONFIGITEM" ]]; then
        echo "$CONFIGITEM"
      fi
    done <<< "$CONFIGFILES"
  )"

  # Get configuration data from files
  CONFIGDATA="$(
    while read -rs CONFIGITEM; do
      if [[ "$CONFIGITEM" ]]; then
        echo "# Configuration file \`${CONFIGITEM}'"
        echo "$(< "$CONFIGITEM")"
      fi
    done <<< "$CONFIGAPPLIED"
  )"

  unset -v CONFIGITEM
  # Apply configuration data
  # shellcheck disable=SC1091 # shellcheck doesn't support this
  . /dev/stdin <<< "$CONFIGDATA"
fi

# Evaluate command line options
get_options "$@"

# Common preflight checks
check_bin xauth xprop xwininfo hostname logger "$GETOPTBIN" "$ZBIN" || exit 1
check_local_display || exit 1

# More preflight checks and main functions
if [[ "$REMCMD" ]]; then
  check_bin socat
  run_rexec
elif ! [[ "$NXLOCATION" ]]; then
  check_bin "$LSOFBIN" "$NXAGENTBIN" || exit 1
  run_agent
else
  check_bin id mkfifo openssl pkill setsid socat uuidgen "$BASHBIN" "$SSHBIN" "$NXPROXYBIN" || exit 1
  run_proxy
fi

#-----------------------------------------------------------------------------
